import os
from multiprocessing.pool import ThreadPool
from bisect import bisect_left
import numpy as np
from tomoscan.io import HDF5File
from ..io.reader import load_images_from_dataurl_dict
from ..utils import check_supported, deprecated_class, get_num_threads
from .ccd import CCDFilter


class FlatFieldArrays:
    """
    A class for flat-field normalization
    """

    # the variable below will be True for the derived class
    # which is taylored for to helical case
    _full_shape = False

    _supported_interpolations = ["linear", "nearest"]

    def __init__(
        self,
        radios_shape: tuple,
        flats,
        darks,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        n_threads=None,
    ):
        """
        Initialize a flat-field normalization process.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where each key is the flat index, and the value is a
            numpy.ndarray of the flat image.
        darks: dict
            Dictionary where each key is the dark index, and the value is a
            numpy.ndarray of the dark image.
        radios_indices: array of int, optional
            Array containing the radios indices in the scan. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to correct flat distortions based on each radio.
        nan_value: float, optional
            Which float value is used to replace nan/inf after flat-field.
        radios_srcurrent: array, optional
            Array with the same shape as radios_indices. Each item contains the synchrotron electric current.
            If not None, normalization with current is applied.
            Please refer to "Notes" for more information on this normalization.
        flats_srcurrent: array, optional
            Array with the same length as "flats". Each item is a measurement of the synchrotron electric current
            for the corresponding flat. The items must be ordered in the same order as the flats indices (`flats.keys()`).
            This parameter must be used along with 'radios_srcurrent'.
            Please refer to "Notes" for more information on this normalization.
        n_threads: int or None, optional
            Number of threads to use for flat-field correction. Default is to use half the threads.

        Important
        ----------
        `flats` and `darks` are expected to be a dictionary with integer keys (the flats/darks indices)
        and numpy array values.
        You can use the following helper functions: `nabu.io.reader.load_images_from_dataurl_dict`
        and `nabu.io.utils.create_dict_of_indices`


        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.

        The normalization with synchrotron electric current is done as follows.
        Let s = sr/sr_max denote the ratio between current and maximum current,
        D be the dark-current frame, and X' be the normalized frame. Then:
          srcurrent_normalization(X) = X' = (X - D)/s + D
          flatfield_normalization(X') = (X' - D)/(F' - D) = (X - D) / (F - D) * sF/sX
        So current normalization boils down to a scalar multiplication after flat-field.
        """
        if self._full_shape:  # noqa: SIM102
            # this is never going to happen in this base class. But in the derived class for helical
            # which needs to keep the full shape
            if radios_indices is not None:
                radios_shape = (len(radios_indices),) + radios_shape[1:]

        self._set_parameters(radios_shape, radios_indices, interpolation, nan_value)
        self._set_flats_and_darks(flats, darks)
        self._precompute_flats_indices_weights()
        self._configure_srcurrent_normalization(radios_srcurrent, flats_srcurrent)
        self.distortion_correction = distortion_correction
        self.n_threads = max(1, get_num_threads(n_threads) // 2)

    def _set_parameters(self, radios_shape, radios_indices, interpolation, nan_value):
        self._set_radios_shape(radios_shape)
        if radios_indices is None:
            radios_indices = np.arange(0, self.n_radios, dtype=np.int32)
        else:
            radios_indices = np.array(radios_indices, dtype=np.int32)
            self._check_radios_and_indices_congruence(radios_indices)

        self.radios_indices = radios_indices
        self.interpolation = interpolation
        check_supported(interpolation, self._supported_interpolations, "Interpolation mode")
        self.nan_value = nan_value
        self._radios_idx_to_pos = dict(zip(self.radios_indices, np.arange(self.radios_indices.size)))

    def _set_radios_shape(self, radios_shape):
        if len(radios_shape) == 2:
            self.radios_shape = (1,) + radios_shape
        elif len(radios_shape) == 3:
            self.radios_shape = radios_shape
        else:
            raise ValueError("Expected radios to have 2 or 3 dimensions")
        n_radios, n_z, n_x = self.radios_shape
        self.n_radios = n_radios
        self.n_angles = n_radios
        self.shape = (n_z, n_x)

    def _set_flats_and_darks(self, flats, darks):
        self._check_frames(flats, "flats", 1, 9999)
        self.n_flats = len(flats)
        self.flats = flats
        self._sorted_flat_indices = sorted(self.flats.keys())

        if self._full_shape:
            # this is never going to happen in this base class. But in the derived class for helical
            # which needs to keep the full shape
            self.shape = flats[self._sorted_flat_indices[0]].shape

        self._flat2arrayidx = dict(zip(self._sorted_flat_indices, np.arange(self.n_flats)))
        self.flats_arr = np.zeros((self.n_flats,) + self.shape, "f")
        for i, idx in enumerate(self._sorted_flat_indices):
            self.flats_arr[i] = self.flats[idx]

        self._check_frames(darks, "darks", 1, 1)
        self.darks = darks
        self.n_darks = len(darks)
        self._sorted_dark_indices = sorted(self.darks.keys())
        self._dark = None

    def _check_frames(self, frames, frames_type, min_frames_required, max_frames_supported):
        n_frames = len(frames)
        if n_frames < min_frames_required:
            raise ValueError("Need at least %d %s" % (min_frames_required, frames_type))
        if n_frames > max_frames_supported:
            raise ValueError(
                "Flat-fielding with more than %d %s is not supported" % (max_frames_supported, frames_type)
            )
        self._check_frame_shape(frames, frames_type)

    def _check_frame_shape(self, frames, frames_type):
        for frame_idx, frame in frames.items():
            if frame.shape != self.shape:
                raise ValueError(
                    "Invalid shape for %s %s: expected %s, but got %s"
                    % (frames_type, frame_idx, str(self.shape), str(frame.shape))
                )

    def _check_radios_and_indices_congruence(self, radios_indices):
        if radios_indices.size != self.n_radios:
            raise ValueError(
                "Expected radios_indices to have length %s = n_radios, but got length %d"
                % (self.n_radios, radios_indices.size)
            )

    def _precompute_flats_indices_weights(self):
        """
        Build two arrays: "indices" and "weights".
        These arrays contain pre-computed information so that the interpolated flat is obtained with

           flat_interpolated = weight_prev * flat_prev  + weight_next * flat_next

        where
           weight_prev, weight_next = weights[2*i], weights[2*i+1]
           idx_prev, idx_next = indices[2*i], indices[2*i+1]
           flat_prev, flat_next = flats[idx_prev], flats[idx_next]

        In words:
          - If a projection has an index between two flats, the equivalent flat is a linear interpolation
            between "previous flat" and "next flat".
          - If a projection has the same index as a flat, only this flat is used for normalization
            (this case normally never occurs, but it's handled in the code)
        """

        def _interp_linear(idx, prev_next):
            if len(prev_next) == 1:  # current index corresponds to an acquired flat
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[prev_next[0]], -1)
            else:
                prev_idx, next_idx = prev_next
                delta = next_idx - prev_idx
                w1 = 1 - (idx - prev_idx) / delta
                w2 = 1 - (next_idx - idx) / delta
                weights = (w1, w2)
                f_idx = (self._flat2arrayidx[prev_idx], self._flat2arrayidx[next_idx])
            return f_idx, weights

        def _interp_nearest(idx, prev_next):
            if len(prev_next) == 1:  # current index corresponds to an acquired flat
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[prev_next[0]], -1)
            else:
                prev_idx, next_idx = prev_next
                idx_to_take = prev_idx if abs(idx - prev_idx) < abs(idx - next_idx) else next_idx
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[idx_to_take], -1)
            return f_idx, weights

        self.flats_idx = np.zeros((self.n_radios, 2), dtype=np.int32)
        self.flats_weights = np.zeros((self.n_radios, 2), dtype=np.float32)
        for i, idx in enumerate(self.radios_indices):
            prev_next = self.get_previous_next_indices(self._sorted_flat_indices, idx)
            if self.interpolation == "nearest":
                f_idx, weights = _interp_nearest(idx, prev_next)
            elif self.interpolation == "linear":
                f_idx, weights = _interp_linear(idx, prev_next)
            # pylint: disable=E0606
            self.flats_idx[i] = f_idx
            self.flats_weights[i] = weights

    # pylint: disable=E1307
    def _configure_srcurrent_normalization(self, radios_srcurrent, flats_srcurrent):
        self.normalize_srcurrent = False
        if radios_srcurrent is None or flats_srcurrent is None:
            return
        radios_srcurrent = np.array(radios_srcurrent)
        if radios_srcurrent.size != self.n_radios:
            raise ValueError(
                "Expected 'radios_srcurrent' to have %d elements but got %d" % (self.n_radios, radios_srcurrent.size)
            )
        flats_srcurrent = np.array(flats_srcurrent)
        if flats_srcurrent.size != self.n_flats:
            raise ValueError(
                "Expected 'flats_srcurrent' to have %d elements but got %d" % (self.n_flats, flats_srcurrent.size)
            )
        self.normalize_srcurrent = True
        self.radios_srcurrent = radios_srcurrent
        self.flats_srcurrent = flats_srcurrent
        self.srcurrent_ratios = np.zeros(self.n_radios, "f")
        # Flats SRCurrent is obtained with "nearest" interp, to emulate an already-done flats SR current normalization
        for i, radio_idx in enumerate(self.radios_indices):
            flat_idx = self.get_nearest_index(self._sorted_flat_indices, radio_idx)
            flat_srcurrent = self.flats_srcurrent[self._flat2arrayidx[flat_idx]]
            self.srcurrent_ratios[i] = flat_srcurrent / self.radios_srcurrent[i]

    @staticmethod
    def get_previous_next_indices(arr, idx):
        pos = bisect_left(arr, idx)
        if pos == len(arr):  # outside range
            return (arr[-1],)
        if arr[pos] == idx:
            return (idx,)
        if pos == 0:
            return (arr[0],)
        return arr[pos - 1], arr[pos]

    @staticmethod
    def get_nearest_index(arr, idx):
        pos = bisect_left(arr, idx)
        if pos == len(arr) or arr[pos] == idx:
            return arr[-1]
        return arr[pos - 1] if idx - arr[pos - 1] < arr[pos] - idx else arr[pos]

    @staticmethod
    def interp(pos, indices, weights, array, slice_y=slice(None, None), slice_x=slice(None, None)):
        """
        Interpolate between two values. The interpolator consists in pre-computed arrays such that

           prev, next = indices[pos]
           w1, w2 = weights[pos]
           interpolated_value = w1 * array[prev] + w2 * array[next]
        """
        prev_idx = indices[pos, 0]
        next_idx = indices[pos, 1]

        if slice_y != slice(None, None) or slice_x != slice(None, None):
            w1 = weights[pos, 0][slice_y, slice_x]
            w2 = weights[pos, 1][slice_y, slice_x]
        else:
            w1 = weights[pos, 0]
            w2 = weights[pos, 1]

        if next_idx == -1:
            val = array[prev_idx]
        else:
            val = w1 * array[prev_idx] + w2 * array[next_idx]
        return val

    def get_flat(self, pos, dtype=np.float32, slice_y=slice(None, None), slice_x=slice(None, None)):
        flat = self.interp(pos, self.flats_idx, self.flats_weights, self.flats_arr, slice_y=slice_y, slice_x=slice_x)
        if flat.dtype != dtype:
            flat = np.ascontiguousarray(flat, dtype=dtype)
        return flat

    def get_dark(self):
        if self._dark is None:
            first_dark_idx = self._sorted_dark_indices[0]
            dark = np.ascontiguousarray(self.darks[first_dark_idx], dtype=np.float32)
            self._dark = dark
        return self._dark

    def remove_invalid_values(self, img):
        if self.nan_value is None:
            return
        invalid_mask = np.logical_not(np.isfinite(img))
        img[invalid_mask] = self.nan_value

    def normalize_radios(self, radios):
        """
        Apply a flat-field normalization, with the current parameters, to a stack
        of radios.
        The processing is done in-place, meaning that the radios content is overwritten.

        Parameters
        -----------
        radios: numpy.ndarray
            Radios chunk
        """
        do_flats_distortion_correction = self.distortion_correction is not None
        dark = self.get_dark()

        def apply_flatfield(i):
            radio_data = radios[i]
            radio_data -= dark
            flat = self.get_flat(i)
            flat = flat - dark
            if do_flats_distortion_correction:
                flat = self.distortion_correction.estimate_and_correct(flat, radio_data)
            np.divide(radio_data, flat, out=radio_data)
            self.remove_invalid_values(radio_data)

        if self.n_threads > 2:
            with ThreadPool(self.n_threads) as tp:
                tp.map(apply_flatfield, range(self.n_radios))
        else:
            for i in range(self.n_radios):
                apply_flatfield(i)

        if self.normalize_srcurrent:
            radios *= self.srcurrent_ratios[:, np.newaxis, np.newaxis]
        return radios

    def normalize_single_radio(
        self, radio, radio_idx, dtype=np.float32, slice_y=slice(None, None), slice_x=slice(None, None)
    ):
        """
        Apply a flat-field normalization to a single projection image.
        """
        dark = self.get_dark()[slice_y, slice_x]
        radio -= dark
        radio_pos = self._radios_idx_to_pos[radio_idx]
        flat = self.get_flat(radio_pos, dtype=dtype, slice_y=slice_y, slice_x=slice_x)
        flat = flat - dark
        if self.distortion_correction is not None:
            flat = self.distortion_correction.estimate_and_correct(flat, radio)
        radio /= flat
        if self.normalize_srcurrent:
            radio *= self.srcurrent_ratios[radio_pos]
        self.remove_invalid_values(radio)
        return radio


FlatField = FlatFieldArrays


@deprecated_class(
    "FlatFieldDataUrls is deprecated since 2024.2.0 and will be removed in a future version", do_print=True
)
class FlatFieldDataUrls(FlatField):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        **chunk_reader_kwargs,
    ):
        """
        Initialize a flat-field normalization process with DataUrls.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where the key is the flat index, and the value is a
            silx.io.DataUrl pointing to the flat.
        darks: dict
            Dictionary where the key is the dark index, and the value is a
            silx.io.DataUrl pointing to the dark.
        radios_indices: array, optional
            Array containing the radios indices. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to correct flat distortions based on each radio.
        nan_value: float, optional
            Which float value is used to replace nan/inf after flat-field.


        Other Parameters
        ----------------
        The other named parameters are passed to ChunkReader(). Please read its
        documentation for more information.

        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.
        """

        flats_arrays_dict = load_images_from_dataurl_dict(flats, **chunk_reader_kwargs)
        darks_arrays_dict = load_images_from_dataurl_dict(darks, **chunk_reader_kwargs)
        super().__init__(
            radios_shape,
            flats_arrays_dict,
            darks_arrays_dict,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
            nan_value=nan_value,
            radios_srcurrent=radios_srcurrent,
            flats_srcurrent=flats_srcurrent,
        )


class PCAFlatsNormalizer:
    """This class implement a flatfield normalization based on a PCA of a series of acquired flatfields.
    The PCA decomposition is handled by a PCAFlatsDecomposer object.

    This implementation was proposed by Jailin C. et al in https://journals.iucr.org/s/issues/2017/01/00/fv5055/

    Code initially written by ID11 @ ESRF staff.
    Jonathan Wright - Implementation based on research paper
    Pedro D. Resende - Added saving and loading from file capabilities

    Jerome Lesaint - Integrated the solution in Nabu.

    """

    def __init__(self, components, dark, mean):
        """Initializes all variables needed to perform the flatfield normalization.

        Parameters
        -----------
        components: ndarray
            The components of the PCA decomposition.
        dark: ndarray
            The dark image. Should be one single 2D image.
        mean: ndarray
            The mean image of the series of flats.
        """
        ones = np.ones_like(components[0], dtype=np.float32)[np.newaxis]  # This comp will account for I0
        self.components = np.concatenate([ones, components], axis=0)
        self.dark = dark
        self.mean = mean
        self.n_threads = max(1, get_num_threads() // 2)
        self._setmask()
        self.ccdfilter = CCDFilter(mean.shape)

    def _form_lsq_matrix(self):
        """This function form the Least Square matrix, based on the flats components and the mask."""
        self.Amat = np.stack([gg for gg in self.g], axis=1)  # JL: this is the matrix for the fit

    def _setmask(self, prop=0.125):
        """Sets the mask to select where the model is going to be fitted.

        Parameters
        ----------
        prop: float, default: 0.125
            The proportion of the image width to take on each side of the image as a mask.

            By default it sets the strips on each side of the frame in the form:
                mask[:, lim:] = True
                mask[:, -lim:] = True
            Where lim = prop * flat.shape[1]

        If you need a custom mask, see update_mask() method.
        """

        lim = int(prop * self.mean.shape[1])
        self.mask = np.zeros(self.mean.shape, dtype=bool)
        self.mask[:, :lim] = True
        self.mask[:, -lim:] = True
        self.g = []
        for component in self.components:
            self.g.append(component[self.mask])

        self._form_lsq_matrix()

    def update_mask(self, mask: np.ndarray):
        """Method to update the mask with a custom mask in the form of a boolean 2+D array.

        Paramters
        =========
        mask: np.ndarray of Boolean;
            The array of boolean allows the selection of the region of the image that will be used to fit against the components.

        It will set the mask, replacing the standard mask created with setmask().
        """
        if mask.dtype == bool:
            self.mask = mask
            self.g = []
            for component in self.components:
                self.g.append(component[self.mask])

            self._form_lsq_matrix()
        else:
            raise TypeError("Not a boolean array. Will keep the default mask")

    def normalize_radios(self, projections, mask=None, prop=0.125):
        """This is to keep the flatfield API in the pipeline."""
        self.correct_stack(projections, mask=mask, prop=prop)

    def correct_stack(
        self,
        projections: np.ndarray,
        mask: np.ndarray = None,
        prop: float = 0.125,
    ):
        """This functions normalizes the stack of projections.

        Performs correction on a stack of projections based on the calculated decomposition.
        The normalizations is done in-place. The previous projections before normalization are lost.

        Parameters
        ----------
        projections: ndarray
            Stack of projections to normalize.

        prop: float (default: {0.125})
            Fraction to mask on the horizontal field of view, assuming vertical rotation axis

        mask: np.ndarray (default: None)
            Custom mask if your data requires it.

        Returns
        -------
        corrected projections: np.ndarray
            Flat field corrected images. Note that the returned projections are exp-transformed to fit the pipeline (e.g. to allow for phase retrieval).
        """
        self.projections = self.ccdfilter.dezinger_correction(projections, self.dark)

        if mask is not None:
            self.update_mask(mask=mask)
        else:
            self._setmask(prop=prop)

        with ThreadPool(self.n_threads) as tp:
            for i, cor, sol in tp.map(self._readcorrect1, range(len(projections))):
                # solution[i] = sol
                projections[i] = np.exp(-cor)

    def _readcorrect1(self, ii):
        """Method to allow parallelization of the normalization."""
        corr, s = self.correctproj(self.projections[ii])
        return ii, corr, s

    def correctproj(self, projection):
        """Performs the correction on one projection of the stack.

        Parameters
        ----------
        projection: np.ndarray, float
            Radiograph from the acquisition stack.

        Returns
        -------
        The fitted projection.
        """
        logp = np.log(projection.astype(np.float32) - self.dark)
        corr = self.mean - logp
        # model to be fitted !!
        return self.fit(corr)

    def fit(self, corr):
        """Fit the (masked) projection to the (masked) components of the PCA decomposition.

        This is for each projection, so worth optimising ...
        """
        y = corr[self.mask]
        solution = np.linalg.lstsq(self.Amat, y, rcond=None)[0]
        correction = np.einsum("ijk,i->jk", self.components, solution)
        return corr - correction, solution


class PCAFlatsDecomposer:
    """This class implements a PCA decomposition of a serie of acquired flatfields.
    The PCA decomposition is used to normalize the projections through a PCAFLatNormalizer object.

    This implementation was proposed by Jailin C. et al in https://journals.iucr.org/s/issues/2017/01/00/fv5055/

    Code initially written by ID11 @ ESRF staff.
    Jonathan Wright - Implementation based on research paper
    Pedro D. Resende - Added saving and loading from file capabilities

    Jerome Lesaint - Integrated the solution in Nabu.
    """

    def __init__(self, flats: np.ndarray, darks: np.ndarray, nsigma=3):
        """

        Parameters
        -----------
        flats: np.ndarray
            A stack of darks corrected flat field images
        darks: np.ndarray
            An image or stack of images of the dark current images of the camera.

        Does the log scaling.
        Subtracts mean and does eigenvector decomposition.
        """
        self.n_threads = max(1, get_num_threads() // 2)
        self.flats = np.empty(flats.shape, dtype=np.float32)

        darks = darks.astype(np.float32)
        if darks.ndim == 3:
            self.dark = np.median(darks, axis=0)
        else:
            self.dark = darks.copy()
        del darks

        self.nsigma = nsigma
        self.ccdfilter = CCDFilter(self.dark.shape)
        self._ccdfilter_and_log(flats)  # Log is taken here (after dezinger)

        self.mean = np.mean(self.flats, axis=0)  # average

        self.flats = self.flats - self.mean  # makes a copy
        self.cov = self.compute_correlation_matrix()
        self.compute_pca()
        self.generate_pca_flats(nsigma=nsigma)  # Default nsigma=3

    def __str__(self):

        return f"PCA decomposition from flat images. \nThere are {self.components} components created at {self.sigma} level."

    def _ccdfilter_and_log(self, flats: np.ndarray):
        """Dezinger (substract dark, apply median filter) and takes log of the flat stack"""

        self.ccdfilter.dezinger_correction(flats, self.dark)

        with ThreadPool(self.n_threads) as tp:
            for i, frame in enumerate(tp.map(np.log, flats)):
                self.flats[i] = frame

    @staticmethod
    def ccij(args):
        """Compute the covariance (img[i]*img[j]).sum() / npixels
        It is a wrapper for threading.
        args == i, j, npixels, imgs
        """

        i, j, NY, imgs = args
        return i, j, np.einsum("ij,ij", imgs[i], imgs[j]) / NY

    def compute_correlation_matrix(self):
        """Computes an (nflats x nflats) correlation matrix"""

        N = len(self.flats)
        CC = np.zeros((N, N), float)
        args = [(i, j, N, self.flats) for i in range(N) for j in range(i + 1)]
        with ThreadPool(self.n_threads) as tp:
            for i, j, result in tp.map(self.ccij, args):
                CC[i, j] = CC[j, i] = result
        return CC

    def compute_pca(self):
        """Gets eigenvectors and eigenvalues and sorts them into order"""

        self.eigenvalues, self.eigenvectors = np.linalg.eigh(
            self.cov
        )  # Not sure why the eigh is needed. eig should be enough.
        order = np.argsort(abs(self.eigenvalues))[::-1]  # high to low
        self.eigenvalues = self.eigenvalues[order]
        self.eigenvectors = self.eigenvectors[:, order]

    def generate_pca_flats(self, nsigma=3):
        """Projects the eigenvectors back into image space.

        Parameters
        ----------
        nsigma: int (default: 3)
        """

        self.sigma = nsigma
        av = abs(self.eigenvalues)
        N = (av > (av[-2] * nsigma)).sum()  # Go for 3 sigma
        self.components = [
            None,
        ] * N

        def calculate(ii):
            calc = np.einsum("i,ijk->jk", self.eigenvectors[:, ii], self.flats)
            norm = (calc**2).sum()
            return ii, calc / np.sqrt(norm)

        with ThreadPool(self.n_threads) as tp:
            for ii, result in tp.map(calculate, range(N)):
                self.components[ii] = result

        # simple gradients
        r, c = self.components[0].shape
        self.components.append(np.outer(np.ones(r), np.linspace(-1 / c, 1 / c, c)))
        self.components.append(np.outer(np.linspace(-1 / r, 1 / r, r), np.ones(c)))

    def save_decomposition(self, path="PCA_flats.h5", overwrite=True, entry="entry0000"):
        """Saves the basic information of a PCA decomposition in view of the normalization of projections.

        Parameters
        ----------
        path: str (default: "PCA_flats.h5")
            Full path to the h5 file you want to save your results. It will overwrite!! Be careful.
        """

        file_exists = os.path.exists(path)
        if overwrite or not file_exists:
            with HDF5File(path, "w") as hout:
                group = hout.create_group(entry)
                group["eigenvalues"] = self.eigenvalues
                group["dark"] = self.dark.astype(np.float32)
                group["p_mean"] = self.mean.astype(np.float32)
                group["p_components"] = np.array(self.components)
                hout.flush()
        else:
            raise OSError(f"The file {path} already exists and you chose to NOT overwrite.")



from math import pi
import numpy as np
from scipy.fft import rfft, irfft
from silx.image.tomography import compute_fourier_filter, get_next_power
from ..processing.padding_base import PaddingBase
from ..utils import check_supported, get_num_threads


class SinoFilter:
    available_filters = [
        "ramlak",
        "shepp-logan",
        "cosine",
        "hamming",
        "hann",
        "tukey",
        "lanczos",
        "hilbert",
    ]

    """
    A class for sinogram filtering.
    It does the following:
      - pad input array
      - Fourier transform each row
      - multiply with a 1D or 2D filter
      - inverse Fourier transform
    """

    available_padding_modes = PaddingBase.supported_modes
    default_extra_options = {"cutoff": 1.0, "fft_threads": 0}  # use all threads by default

    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        crop_filtered_data=True,
        extra_options=None,
    ):
        """
        Initialize a SinoFilter instance.

        Parameters
        ----------
        sino_shape: tuple
            Shape of sinogram, in the form (n_angles, detector_width) or (n_sinos, n_angles, detector_width)
        filter_name: str, optional
            Name of the filter. Default is ram-lak.
        padding_mode: str, optional
            How to pad the data prior to filtering. Default is zero-padding, corresponding to linear convolution with the filter kernel.
            In practice this value is often set to "edges" for interior tomography.
        crop_filtered_data: bool, optional
            Whether to crop the final, filtered sinogram. Default is True. See notes below.
        extra_options: dict, optional
            Dictionary of advanced extra options.

        Notes
        -----
        Sinogram filtering done in the Filtered Back-Projection (FBP) method consists, in theory, in applying a high-pass filter
        to the sinogram prior to backprojection. This high-pass filter is normally the Ramachandran-Lakshminarayanan (Ram-Lak) filter
        yielding a close-to-ideal reconstruction (see Natterer's "Mathematical methods in image reconstruction").
        As the filter kernel has a large extent in spatial domain, it's best performed in Fourier domain via the Fourier-convolution theorem.
        Filtering in Fourier domain should be done with a data padded to at least twice its size.
        Zero-padding should be used for mathematical correctness (so that multiplication in Fourier domain corresponds to an actual linear convolution).
        However if the sinogram does not decay to "zero" near the edges (i.e in interior tomography), padding with zeros usually gives artefacts after filtering.
        In this case, padding with edges is preferred (corresponding to a convolution with the "edges" extension mode).

        After inverse Fourier transform, the (padded and filtered) data is cropped back to its original size.
        In some cases, it's preferable to keep the data un-cropped for further processing.
        """

        self._init_extra_options(extra_options)
        self._set_padding_mode(padding_mode)
        self._calculate_shapes(sino_shape, crop_filtered_data)
        self._init_fft()
        self._allocate_memory()
        self._compute_filter(filter_name)

    def _init_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _set_padding_mode(self, padding_mode):
        # Compat.
        if padding_mode == "edges":
            padding_mode = "edge"
        if padding_mode == "zeros":
            padding_mode = "constant"
        #
        check_supported(padding_mode, self.available_padding_modes, "padding mode")
        self.padding_mode = padding_mode

    def _calculate_shapes(self, sino_shape, crop_filtered_data):
        self.ndim = len(sino_shape)
        if self.ndim == 2:
            n_angles, dwidth = sino_shape
            n_sinos = 1
        elif self.ndim == 3:
            n_sinos, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Invalid sinogram number of dimensions")
        self.sino_shape = sino_shape
        self.n_angles = n_angles
        self.dwidth = dwidth
        # Make sure to use int() here, otherwise pycuda/pyopencl will crash in some cases
        self.dwidth_padded = int(get_next_power(2 * self.dwidth))
        self.sino_padded_shape = (n_angles, self.dwidth_padded)
        if self.ndim == 3:
            self.sino_padded_shape = (n_sinos,) + self.sino_padded_shape
        sino_f_shape = list(self.sino_padded_shape)
        sino_f_shape[-1] = sino_f_shape[-1] // 2 + 1
        self.sino_f_shape = tuple(sino_f_shape)
        self.pad_left = (self.dwidth_padded - self.dwidth) // 2
        self.pad_right = self.dwidth_padded - self.dwidth - self.pad_left

        self.crop_filtered_data = crop_filtered_data
        self.output_shape = self.sino_shape
        if not (self.crop_filtered_data):
            self.output_shape = self.sino_padded_shape

    def _init_fft(self):
        pass

    def _allocate_memory(self):
        pass

    def set_filter(self, h_filt, normalize=True):
        """
        Set a filter for sinogram filtering.

        Parameters
        ----------
        h_filt: numpy.ndarray
            Array containing the filter. Each line of the sinogram will be filtered with
            this filter. It has to be the Real-to-Complex Fourier Transform
            of some real filter, padded to 2*sinogram_width.
        normalize: bool or float, optional
            Whether to normalize (multiply) the filter with pi/num_angles.
        """
        if h_filt.size != self.sino_f_shape[-1]:
            raise ValueError(
                """
                Invalid filter size: expected %d, got %d.
                Please check that the filter is the Fourier R2C transform of
                some real 1D filter.
                """
                % (self.sino_f_shape[-1], h_filt.size)
            )
        if not (np.iscomplexobj(h_filt)):
            print("Warning: expected a complex Fourier filter")
        self.filter_f = h_filt.copy()
        if normalize:
            self.filter_f *= pi / self.n_angles
        self.filter_f = self.filter_f.astype(np.complex64)

    def _compute_filter(self, filter_name):
        self.filter_name = filter_name or "ram-lak"
        # TODO add this one into silx
        if self.filter_name == "hilbert":
            freqs = np.fft.fftfreq(self.dwidth_padded)
            filter_f = 1.0 / (2 * pi * 1j) * np.sign(freqs)
        #
        else:
            filter_f = compute_fourier_filter(
                self.dwidth_padded,
                self.filter_name,
                cutoff=self.extra_options["cutoff"],
            )
        filter_f = filter_f[: self.dwidth_padded // 2 + 1]  # R2C
        self.set_filter(filter_f, normalize=True)

    def _check_array(self, arr):
        if arr.dtype != np.float32:
            raise ValueError("Expected data type = numpy.float32")
        if arr.shape != self.sino_shape:
            raise ValueError("Expected sinogram shape %s, got %s" % (self.sino_shape, arr.shape))

    def filter_sino(self, sino, output=None):
        """
        Perform the sinogram siltering.

        Parameters
        ----------
        sino: array
            Input sinogram (2D or 3D)
        output: array, optional
            Output array.
        """
        self._check_array(sino)
        # sino_padded = np.pad(
        #     sino, ((0, 0), (0, self.dwidth_padded - self.dwidth)), mode=self.padding_mode
        # )  # pad with a FFT-friendly layout
        sino_padded = np.pad(sino, ((0, 0), (self.pad_left, self.pad_right)), mode=self.padding_mode)
        sino_padded_f = rfft(sino_padded, axis=1, workers=get_num_threads(self.extra_options["fft_threads"]))
        sino_padded_f *= self.filter_f
        sino_filtered = irfft(sino_padded_f, axis=1, workers=get_num_threads(self.extra_options["fft_threads"]))

        if output is None:
            if not (self.crop_filtered_data):
                # No need to allocate extra memory here
                return sino_filtered
            res = np.zeros(self.output_shape, dtype=np.float32)
        else:
            res = output

        if self.crop_filtered_data:
            # res[:] = sino_filtered[..., : self.dwidth]  # pylint: disable=E1126 # ?!
            res[:] = sino_filtered[..., self.pad_left : -self.pad_right]  # pylint: disable=E1126 # ?!
        else:
            res[:] = sino_filtered[:]

        return res

    __call__ = filter_sino


def filter_sinogram(
    sinogram,
    padded_width,
    filter_name="ramlak",
    padding_mode="constant",
    normalize=True,
    filter_cutoff=1.0,
    crop_filtered_data=True,
    **padding_kwargs,
):
    """
    Simple function to filter sinogram.

    Parameters
    ----------
    sinogram: numpy.ndarray
        Sinogram, two dimensional array with shape (n_angles, sino_width)
    padded_width: int
        Width to use for padding. Must be greater than sinogram width (i.e than sinogram.shape[-1])
    filter_name: str, optional
        Which filter to use. Default is ramlak (roughly equivalent to abs(nu) in frequency domain)
    padding_mode: str, optional
        Which padding mode to use. Default is zero-padding.
    normalize: bool, optional
        Whether to multiply the filtered sinogram with pi/n_angles
    filter_cutoff: float, optional
        frequency cutoff for filter
    """
    n_angles, width = sinogram.shape

    # Initially, padding was done this way
    # sinogram_padded = np.pad(sinogram, ((0, 0), (0, padded_width - width)), mode=padding_mode, **padding_kwargs)

    #
    pad_left = (padded_width - width) // 2
    pad_right = padded_width - width - pad_left
    sinogram_padded = np.pad(sinogram, ((0, 0), (pad_left, pad_right)), mode=padding_mode, **padding_kwargs)
    #

    fourier_filter = compute_fourier_filter(padded_width, filter_name, cutoff=filter_cutoff)
    if normalize:
        fourier_filter *= np.pi / n_angles
    fourier_filter = fourier_filter[: padded_width // 2 + 1]  # R2C
    sino_f = rfft(sinogram_padded, axis=1)
    sino_f *= fourier_filter
    sino_filtered = irfft(sino_f, axis=1)
    if crop_filtered_data:
        # sino_filtered = sino_filtered[:, :width]  # pylint: disable=E1126 # ?!
        sino_filtered = sino_filtered[:, pad_left:-pad_right]  # pylint: disable=E1126 # ?!
    return sino_filtered



import os
import numpy as np
import pytest

import h5py
from nabu.testutils import utilstest
from nabu.preproc.flatfield import (
    PCAFlatsDecomposer,
    PCAFlatsNormalizer,
)


@pytest.fixture(scope="class")
def bootstrap_pcaflats(request):
    cls = request.cls
    # TODO: these tolerances for having the tests passed should be tighter.
    # Discrepancies between id11 code and nabu code are still mysterious.
    cls.mean_abs_tol = 1e-1
    cls.comps_abs_tol = 1e-2
    cls.projs, cls.flats, cls.darks = get_pcaflats_data("test_pcaflats.npz")
    cls.raw_projs = cls.projs.copy()  # Needed because flat correction is done inplace.
    ref_data = get_pcaflats_refdata("ref_pcaflats.npz")
    cls.mean = ref_data["mean"]
    cls.components_3 = ref_data["components_3"]
    cls.components_15 = ref_data["components_15"]
    cls.dark = ref_data["dark"]
    cls.normalized_projs_3 = ref_data["normalized_projs_3"]
    cls.normalized_projs_15 = ref_data["normalized_projs_15"]
    cls.normalized_projs_custom_mask = ref_data["normalized_projs_custom_mask"]
    cls.test_normalize_projs_custom_prop = ref_data["normalized_projs_custom_prop"]

    cls.h5_filename_3 = get_h5_pcaflats("pcaflat_3.h5")
    cls.h5_filename_15 = get_h5_pcaflats("pcaflat_15.h5")


def get_pcaflats_data(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    data = np.load(dataset_downloaded_path)
    projs = data["projs"].astype(np.float32)
    flats = data["flats"].astype(np.float32)
    darks = data["darks"].astype(np.float32)

    return projs, flats, darks


def get_h5_pcaflats(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)

    return dataset_downloaded_path


def get_pcaflats_refdata(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    data = np.load(dataset_downloaded_path)

    return data


def get_decomposition(filename):
    with h5py.File(filename, "r") as f:
        # Load the dataset
        p_comps = f["entry0000/p_components"][()]
        p_mean = f["entry0000/p_mean"][()]
        dark = f["entry0000/dark"][()]
    return p_comps, p_mean, dark


@pytest.mark.usefixtures("bootstrap_pcaflats")
class TestPCAFlatsDecomposer:
    def test_decompose_flats(self):
        # Build 3-sigma basis
        pca = PCAFlatsDecomposer(self.flats, self.darks, nsigma=3)
        message = f"Found a discrepency between computed mean flat and reference."
        assert np.allclose(self.mean, pca.mean, atol=self.mean_abs_tol), message
        message = f"Found a discrepency between computed components and reference ones if nsigma=3."
        assert np.allclose(self.components_3, np.array(pca.components), atol=self.comps_abs_tol), message

        # Build 1.5-sigma basis
        pca = PCAFlatsDecomposer(self.flats, self.darks, nsigma=1.5)
        message = f"Found a discrepency between computed components and reference ones, if nsigma=1.5."
        assert np.allclose(self.components_15, np.array(pca.components), atol=self.comps_abs_tol), message

    def test_save_load_decomposition(self):
        pca = PCAFlatsDecomposer(self.flats, self.darks, nsigma=3)
        tmp_path = os.path.join(os.path.dirname(self.h5_filename_3), "PCA_Flats.h5")
        pca.save_decomposition(path=tmp_path)
        p_comps, p_mean, dark = get_decomposition(tmp_path)
        message = f"Found a discrepency between saved and loaded mean flat."
        assert np.allclose(self.mean, p_mean, atol=self.mean_abs_tol), message
        message = f"Found a discrepency between saved and loaded components if nsigma=3."
        assert np.allclose(self.components_3, p_comps, atol=self.comps_abs_tol), message
        message = f"Found a discrepency between saved and loaded dark."
        assert np.allclose(self.dark, dark, atol=self.comps_abs_tol), message
        # Clean up
        if os.path.exists(tmp_path):
            os.remove(tmp_path)


@pytest.mark.usefixtures("bootstrap_pcaflats")
class TestPCAFlatsNormalizer:
    def test_load_pcaflats(self):
        """Tests that the structure of the output PCAFlat h5 file is correct."""
        p_comps, p_mean, dark = get_decomposition(self.h5_filename_3)
        # Check the shape of the loaded data
        assert p_comps.shape[1:] == p_mean.shape
        assert p_comps.shape[1:] == dark.shape

    def test_normalize_projs(self):
        p_comps, p_mean, dark = get_decomposition(self.h5_filename_3)
        pca = PCAFlatsNormalizer(p_comps, dark, p_mean)
        projs = self.raw_projs.copy()
        pca.normalize_radios(projs)
        assert np.allclose(projs, self.normalized_projs_3, atol=1e-2)
        p_comps, p_mean, dark = get_decomposition(self.h5_filename_15)
        pca = PCAFlatsNormalizer(p_comps, dark, p_mean)
        projs = self.raw_projs.copy()
        pca.normalize_radios(projs)
        assert np.allclose(projs, self.normalized_projs_15, atol=1e-2)

    def test_use_custom_mask(self):
        mask = np.zeros(self.mean.shape, dtype=bool)
        mask[:, :10] = True
        mask[:, -10:] = True
        p_comps, p_mean, dark = get_decomposition(self.h5_filename_3)

        pca = PCAFlatsNormalizer(p_comps, dark, p_mean)
        projs = self.raw_projs.copy()
        pca.normalize_radios(projs, mask=mask)
        assert np.allclose(projs, self.normalized_projs_custom_mask, atol=1e-2)

    def test_change_mask_prop(self):
        p_comps, p_mean, dark = get_decomposition(self.h5_filename_3)
        pca = PCAFlatsNormalizer(p_comps, dark, p_mean)
        projs = self.raw_projs.copy()
        pca.normalize_radios(projs, prop=0.05)
        assert np.allclose(projs, self.test_normalize_projs_custom_prop, atol=1e-2)



"""
Double-flatfield:
  - Compute the average of all projections, which gives one resulting image
  - Apply some filter to this image (DFF)
  - Subtract or divide this image from all the projections
"""

from os import path
from silx.io.url import DataUrl
from silx.io.dictdump import h5todict

from nabu.io.utils import get_first_hdf5_entry

from ...utils import is_writeable
from ...app.double_flatfield import DoubleFlatFieldChunks
from ...resources.nxflatfield import data_url_exists

rel_file_path_template = "{scan_name}_dff.h5"
data_path_template = "{entry}/double_flatfield"


def get_possible_dff_urls(dataset_info, user_dir, output_dir):
    """
    See nabu.resources.nxflatfield.get_frame_possible_urls
    """
    entry = dataset_info.hdf5_entry or ""

    def make_dataurl(dirname):
        file_path = path.join(
            dirname,
            rel_file_path_template.format(scan_name=dataset_info.scan_basename),
        )
        return DataUrl(
            file_path=file_path,
            data_path=data_path_template.format(entry=entry),
            scheme="silx",
        )

    urls = {"user": None, "dataset": None, "output": None}

    if user_dir is not None:
        urls["user"] = make_dataurl(user_dir)
    urls["dataset"] = make_dataurl(dataset_info.scan_dirname)
    if output_dir is not None:
        urls["output"] = make_dataurl(output_dir)

    return urls


def compute_and_save_dff(dataset_info, possible_urls, dff_options):
    if possible_urls["user"] is not None:
        dff_output_file = possible_urls["user"].file_path()
    elif is_writeable(path.dirname(possible_urls["dataset"].file_path())):
        dff_output_file = possible_urls["dataset"].file_path()
    else:
        dff_output_file = possible_urls["output"].file_path()

    dataset_info.logger.info("Computing double flatfield")
    dff = DoubleFlatFieldChunks(
        None,
        dff_output_file,
        dataset_info=dataset_info,
        chunk_size=dff_options.get("chunk_size", 100),
        sigma=dff_options.get("dff_sigma", None),
        do_flatfield=dff_options.get("do_flatfield", True),
        logger=dataset_info.logger,
    )
    dff_image = dff.compute_double_flatfield()
    return dff.write_double_flatfield(dff_image)


def check_existing_dff(dff_url, dff_options, logger):
    # Check that the DFF exists at the given DataUrl, and that its configuration matches the wanted config
    # Return the DFF file path
    if not (data_url_exists(dff_url)):
        raise ValueError("DFF file not found:", dff_url)

    fname = dff_url.file_path()
    entry = get_first_hdf5_entry(fname)
    dff_file_options = h5todict(fname, path=entry + "/double_flatfield/configuration", asarray=False)

    ff_file = dff_file_options.get("do_flatfield", True)
    ff_user = dff_options.get("do_flatfield", True)
    # Use "==" instead of "is" here, as h5todict() will return something like numpy.True_ instead of True
    if ff_file != ff_user:
        msg = "DFF was computed with flatfield=%s, but you asked flatfield=%s" % (ff_file, ff_user)
        logger.error(msg)
        return False

    # Use this because h5todict() returns str("None") instead of None
    def _correct_none(x):
        if x in [None, "None"]:
            return None
        return x

    sigma_file = _correct_none(dff_file_options.get("dff_sigma", None))
    sigma_user = _correct_none(dff_options.get("dff_sigma", None))
    if sigma_file != sigma_user:
        msg = "DFF was computed with dff_sigma=%s, but you asked dff_sigma=%s" % (sigma_file, sigma_user)
        logger.error(msg)
        return False

    return fname


# pylint: disable=E1136
def get_double_flatfield(dataset_info, mode, output_dir=None, darks_flats_dir=None, dff_options=None):
    """
    See nabu.resources.nxflatfield.update_dataset_info_flats_darks for the logic
    """
    if mode is False:
        return
    dff_options = dff_options or {}

    possible_urls = get_possible_dff_urls(dataset_info, darks_flats_dir, output_dir)

    if mode == "force-compute":
        return compute_and_save_dff(dataset_info, possible_urls, dff_options)

    def _can_load_from(folder_type):
        if possible_urls.get(folder_type, None) is None:
            return False
        return data_url_exists(possible_urls[folder_type])

    where_to_load_from = None
    if possible_urls["user"] is not None and _can_load_from("user"):
        where_to_load_from = "user"
    elif _can_load_from("dataset"):
        where_to_load_from = "dataset"
    elif _can_load_from("output"):
        where_to_load_from = "output"

    if where_to_load_from is None:
        if mode == "force-load":
            raise ValueError("Could not load double-flatfield file (using 'force-load')")
        else:
            return compute_and_save_dff(dataset_info, possible_urls, dff_options)

    fname = check_existing_dff(possible_urls[where_to_load_from], dff_options, dataset_info.logger)
    if fname is False:
        if mode == "force-load":
            raise ValueError("Could not load double-flatfield file (using 'force-load'): wrong configuration")
        return compute_and_save_dff(dataset_info, possible_urls, dff_options)
    return fname

    # One possible corner case: if mode == "force-load" and darks_flats_dir is not None (but the actual folder is empty)
    # then nabu will load a DFF found elsewhere (if any). We might want to raise an error instead.



from math import cos, sin
import numpy as np
from nabu.utils import search_sorted
from nabu.estimation.motion import MotionEstimation
from nabu.estimation.translation import estimate_shifts
from nabu.testutils import get_data, __do_long_tests__
from scipy.ndimage import shift
import pytest

try:
    import astra
except (ImportError, RuntimeError):
    astra = None


def test_estimate_shifts():
    image = get_data("chelsea.npz")["data"]
    n_tests = 10
    shift_min = -50
    shift_max = 50
    max_tol = 0.2  # in pixel unit
    shifts = np.random.rand(n_tests, 2) * (shift_max - shift_min) - shift_min
    for s in shifts:
        image_shifted = shift(image, s)
        estimated_shifts = estimate_shifts(image, image_shifted)
        abs_diff = np.abs(np.array(estimated_shifts) - np.array(s))
        assert np.all(abs_diff < max_tol), "Wrong estimation for shifts=%s : got %s" % (s, estimated_shifts)


def _bootstrap_test_motion_combined(cls):
    cls._fdesc = get_data("motion/test_correct_motion_combined.npz")
    cls.projs = cls._fdesc["projs"]
    cls.angles = cls._fdesc["angles_rad"]
    cls.shifts = cls._fdesc["shifts"]
    cls.cor = -2.81  # center of rotation in pixel units, (relative to the middle of the detector)
    if __do_long_tests__:
        cls._fdesc = get_data("motion/test_correct_motion_horizontal.npz")
        cls.projs_horiz = cls._fdesc["projs"]
        cls.angles_horiz = cls._fdesc["angles_rad"]
        cls.shifts_horiz = cls._fdesc["shifts"]
        cls._fdesc = get_data("motion/test_correct_motion_vertical.npz")
        cls.projs_vertic = cls._fdesc["projs"]
        cls.angles_vertic = cls._fdesc["angles_rad"]
        cls.shifts_vertic = cls._fdesc["shifts"]


@pytest.fixture(scope="class")
def bootstrap_test_motion_combined(request):
    cls = request.cls
    _bootstrap_test_motion_combined(cls)


@pytest.mark.usefixtures("bootstrap_test_motion_combined")
class TestMotionEstimation:

    def _test_estimate_motion_360(
        self,
        projs,
        angles,
        reference_shifts,
        deg=2,
        deg_z=2,
        tol_x=1e-5,
        tol_y=1e-5,
        tol_z=1e-5,
        estimate_cor=False,
        verbose=True,
    ):
        """
        Test MotionEstimation using pairs of projs
        """
        n_a = projs.shape[0]
        projs1 = projs[: n_a // 2]
        projs2 = projs[n_a // 2 :]
        angles1 = angles[: n_a // 2]
        angles2 = angles[n_a // 2 :]
        estimator = MotionEstimation(projs1, projs2, angles1, angles2, shifts_estimator="DetectorTranslationAlongBeam")
        estimated_shifts_v = estimator.estimate_vertical_motion(degree=deg_z)
        cor = None if estimate_cor else self.cor
        estimated_shifts_h, cor_est = estimator.estimate_horizontal_motion(degree=deg, cor=cor)

        shifts = reference_shifts
        shifts_xy = -shifts[:, :2]
        shifts_z = -shifts[:, 2]

        # mind the signs, nabu vs RTK geometry
        abs_diff_x = np.max(np.abs(estimated_shifts_h[:, 0] - shifts_xy[:360, 0]))
        abs_diff_y = np.max(np.abs(estimated_shifts_h[:, 1] - shifts_xy[:360, 1]))
        abs_diff_z = np.max(np.abs(estimated_shifts_v - shifts_z[:360]))

        if verbose:
            estimator.plot_detector_shifts(cor=cor)
            estimator.plot_movements(cor=cor, angles_rad=angles, gt_xy=shifts_xy, gt_z=shifts_z)

        assert abs_diff_x < tol_x, "Wrong x-movement estimation (estimate_cor=%s)" % estimate_cor
        assert abs_diff_y < tol_y, "Wrong y-movement estimation (estimate_cor=%s)" % estimate_cor
        assert abs_diff_z < tol_z, "Wrong z-movement estimation (estimate_cor=%s)" % estimate_cor

    def test_estimate_motion_360(self, verbose=False):
        """
        Test with a synthetic dataset that underwent horizontal AND vertical translations
        """
        params_dict = {
            "estimate_cor": [False, True],
            "tol_x": [0.05, 0.1],
            "tol_y": [0.05, 0.2],
            "tol_z": [0.005, 0.01],
        }
        params_names = params_dict.keys()
        params_values = np.array(list(params_dict.values()), dtype="O")
        for i in range(params_values.shape[1]):
            params = {k: v for k, v in zip(params_names, params_values[:, i].tolist())}
            self._test_estimate_motion_360(
                self.projs[:360],
                self.angles[:360],
                self.shifts[:360],
                verbose=verbose,
                **params,
            )

    def test_estimate_motion_360_return(self, verbose=False):
        return_angles = np.deg2rad([360, 270, 180, 90, 0])
        outward_angles_indices = np.array([search_sorted(self.angles[:360], ra) for ra in return_angles])
        outward_angles = self.angles[outward_angles_indices]
        projs1 = self.projs[outward_angles_indices]
        projs2 = self.projs[-5:]

        estimator = MotionEstimation(
            projs1,
            projs2,
            outward_angles,
            return_angles,
            indices1=outward_angles_indices,
            indices2=np.arange(self.projs.shape[0])[-5:],
            n_projs_tot=self.projs.shape[0],
        )
        estimated_shifts_z = estimator.estimate_vertical_motion(degree=2)
        estimated_shifts_xy, c = estimator.estimate_horizontal_motion(degree=2, cor=self.cor)

        shifts = self.shifts
        gt_xy = -shifts[:, :2][:-5]
        gt_z = -shifts[:, 2][:-5]

        if verbose:
            estimator.plot_detector_shifts(cor=self.cor)
            estimator.plot_movements(cor=self.cor, angles_rad=self.angles[:-5], gt_xy=gt_xy, gt_z=gt_z)

        estimated_shifts_xy = estimator.apply_fit_horiz(angles=self.angles[:-5])
        estimated_shifts_z = estimator.apply_fit_vertic(angles=self.angles[:-5])

        abs_diff_x = np.max(np.abs(estimated_shifts_xy[:360, 0] - gt_xy[:360, 0]))
        abs_diff_y = np.max(np.abs(estimated_shifts_xy[:360, 1] - gt_xy[:360, 1]))
        abs_diff_z = np.max(np.abs(estimated_shifts_z[:360] - gt_z[:360]))

        max_fit_err_vu = estimator.get_max_fit_error(cor=self.cor)
        assert (
            max_fit_err_vu[0] < 0.7
        ), "Max difference between detector_v_shifts and fit(detector_v_shifts) is too high"  # can't do better for z estimation ?!
        assert (
            max_fit_err_vu[1] < 0.2
        ), "Max difference between detector_u_shifts and fit(detector_u_shifts) is too high"

        assert np.max(abs_diff_x) < 0.3, "Wrong x-movement estimation"
        assert np.max(abs_diff_y) < 0.5, "Wrong y-movement estimation"
        assert np.max(abs_diff_z) < 0.5, "Wrong z-movement estimation"


@pytest.fixture(scope="class")
def bootstrap_test_motion_estimation2(request):
    cls = request.cls
    cls.volume = get_data("motion/mri_volume_subsampled.npy")


def _create_translations_vector(a, alpha, beta):
    return alpha * a**2 + beta * a


def project_volume(vol, angles, tx, ty, tz, cor=0, orig_det_dist=0):
    """
    Forward-project a volume with translations (tx, ty, tz) of the sample
    """
    n_y, n_x, n_z = vol.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z)

    det_row_count = n_z
    det_col_count = max(n_y, n_x)

    vectors = np.zeros((len(angles), 12))
    for i in range(len(angles)):
        theta = angles[i]
        # ray direction
        vectors[i, 0] = sin(theta)
        vectors[i, 1] = -cos(theta)
        vectors[i, 2] = 0

        # center of detector
        vectors[i, 3] = (
            (cos(theta) ** 2) * tx[i] + cos(theta) * sin(theta) * ty[i] - orig_det_dist * sin(theta) + cos(theta) * cor
        )
        vectors[i, 4] = (
            sin(theta) * cos(theta) * tx[i] + (sin(theta) ** 2) * ty[i] + orig_det_dist * cos(theta) + sin(theta) * cor
        )
        vectors[i, 5] = tz[i]

        # vector from detector pixel (0,0) to (0,1)
        vectors[i, 6] = cos(theta)  # uX
        vectors[i, 7] = sin(theta)  # uY
        vectors[i, 8] = 0  # uZ

        # vector from detector pixel (0,0) to (1,0)
        vectors[i, 9] = 0
        vectors[i, 10] = 0
        vectors[i, 11] = 1

    proj_geom = astra.create_proj_geom("parallel3d_vec", det_row_count, det_col_count, vectors)

    sinogram_id, sinogram = astra.create_sino3d_gpu(vol, proj_geom, vol_geom)
    return sinogram


def check_motion_estimation(
    motion_estimator,
    angles,
    cor,
    gt_xy,
    gt_z,
    fit_error_shifts_tol_vu=(1e-5, 1e-5),
    fit_error_det_tol_vu=(1e-5, 1e-5),
    fit_error_tol_xyz=(1e-5, 1e-5, 1e-5),
    fit_error_det_all_angles_tol_vu=(1e-5, 1e-5),
):
    """
    Assess the fit quality of a MotionEstimation object, given known ground truth sample movements.

    Parameters
    ----------
    motion_estimator: MotionEstimation
        MotionEstimation object
    angles: numpy.ndarray
        Array containing rotation angles in radians. For 180° scan it must not contain the "return projections".
    cor: float
        Center of rotation
    gt_xy: numpy.ndarray
        Ground-truth sample movements in (x, y). Shape (n_angles, 2)
    gt_z: numpy.ndarray
        Ground-truth sample movements in z. Shape (n_angles,)
    fit_error_shifts_tol_vu: tuple of float
        Maximum error tolerance when assessing the fit of radios movements,
        i.e see how well the cross-correlation shifts between pairs of radios  are fitted
    fit_error_det_tol_vu: tuple of float
        Maximum error tolerance when assessing the sample movements fit on used angles, projected on the detector
    fit_error_tol_xyz: tuple of float
        Maximum error tolerance when assessing the fit of sample movements, in the sample domain
    fit_error_det_all_angles_tol_vu: tuple of float
        Maximum error tolerance when assessing the sample movements fit on ALL angles, projected on the detector
    """

    is_return = motion_estimator.is_return
    n_angles = angles.size
    outward_angles = motion_estimator.angles1
    return_angles = motion_estimator.angles1
    ma = lambda x: np.max(np.abs(x))
    outward_angles_indices = motion_estimator.indices1

    # (1) Check the fit in the detector domain, wrt measured shifts between pair of projs
    # ------------------------------------------------------------------------------------
    for fit_err, fit_max_error_tol, coord_name in zip(
        motion_estimator.get_max_fit_error(cor), fit_error_shifts_tol_vu, ["v", "u"]
    ):
        assert (
            fit_err < fit_max_error_tol
        ), f"Max error between (measured detector {coord_name}-shifts) and (estimated fit) is too high: {fit_err} > {fit_max_error_tol}"

    # (2) Check the fit wrt the ground truth motion projected onto detector
    # ----------------------------------------------------------------------
    shifts_vu = motion_estimator.apply_fit_to_get_detector_displacements(cor)
    if is_return:
        gt_shifts_vu_1 = motion_estimator.convert_sample_motion_to_detector_shifts(
            gt_xy[outward_angles_indices], gt_z[outward_angles_indices], outward_angles, cor=cor
        )
        gt_shifts_vu_2 = motion_estimator.convert_sample_motion_to_detector_shifts(
            gt_xy[n_angles:], gt_z[n_angles:], return_angles, cor=cor
        )
    else:
        gt_shifts_vu_1 = motion_estimator.convert_sample_motion_to_detector_shifts(
            gt_xy[: n_angles // 2], gt_z[: n_angles // 2], outward_angles, cor=cor
        )
        gt_shifts_vu_2 = motion_estimator.convert_sample_motion_to_detector_shifts(
            gt_xy[n_angles // 2 :], gt_z[n_angles // 2 :], outward_angles, cor=cor
        )
    gt_shifts_vu = gt_shifts_vu_2 - gt_shifts_vu_1
    if not (is_return):
        gt_shifts_vu[:, 1] += (
            2 * cor
        )  # when comparing pairs of opposits projs, delta_u = u(theta+) - u(theta) + 2*cor !

    err_max_vu = np.max(np.abs(gt_shifts_vu - shifts_vu), axis=0)
    assert (
        err_max_vu[0] < fit_error_det_tol_vu[0]
    ), f"Max difference of fit(used_angles) on 'v' coordinate is too high: {err_max_vu[0]} > {fit_error_det_tol_vu[0]}"
    assert (
        err_max_vu[1] < fit_error_det_tol_vu[1]
    ), f"Max difference of fit(used_angles) on 'u'' coordinate is too high: {err_max_vu[1]} > {fit_error_det_tol_vu[1]}"

    # (3) Check the fit wrt ground truth motion of the sample
    # This is less precise when using only return projections,
    # because the movements were estimated from the detector shifts between a few pairs of projections
    # Here apply the shift on all angles (above was only on the angles used to calculate the fit)
    # --------------------------------------------------------------------------------------------
    txy_est_all_angles = motion_estimator.apply_fit_horiz(angles=angles)
    tz_est_all_angles = motion_estimator.apply_fit_vertic(angles=angles)
    err_max_x = ma(txy_est_all_angles[:, 0] - gt_xy[:n_angles, 0])
    err_max_y = ma(txy_est_all_angles[:, 1] - gt_xy[:n_angles, 1])
    err_max_z = ma(tz_est_all_angles - gt_z[:n_angles])
    assert (
        err_max_x < fit_error_tol_xyz[0]
    ), f"Max error for x coordinate is too high: {err_max_x} > {fit_error_tol_xyz[0]}"
    assert (
        err_max_y < fit_error_tol_xyz[1]
    ), f"Max error for y coordinate is too high: {err_max_y} > {fit_error_tol_xyz[1]}"
    assert (
        err_max_z < fit_error_tol_xyz[2]
    ), f"Max error for z coordinate is too high: {err_max_z} > {fit_error_tol_xyz[2]}"

    # (4) Check the fit wrt ground truth motion, projected onto the detector, for all angles
    # ---------------------------------------------------------------------------------------
    estimated_shifts_vu_all_angles = motion_estimator.convert_sample_motion_to_detector_shifts(
        txy_est_all_angles, tz_est_all_angles, angles, cor=cor
    )
    gt_shifts_vu_all_angles = motion_estimator.convert_sample_motion_to_detector_shifts(
        gt_xy[:n_angles], gt_z[:n_angles], angles, cor=cor
    )
    err_max_vu_all_angles = np.max(np.abs(estimated_shifts_vu_all_angles - gt_shifts_vu_all_angles), axis=0)
    assert (
        err_max_vu_all_angles[0] < fit_error_det_all_angles_tol_vu[0]
    ), f"Max error of fit(all_angles) on 'v' coordinate is too high: {err_max_vu_all_angles[0]} > {fit_error_det_all_angles_tol_vu[0]}"
    assert (
        err_max_vu_all_angles[1] < fit_error_det_all_angles_tol_vu[1]
    ), f"Max error  of fit(all_angles) on 'u' coordinate is too high: {err_max_vu_all_angles[1]} > {fit_error_det_all_angles_tol_vu[1]}"


@pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1 for this test")
@pytest.mark.skipif(astra is None, reason="Need astra for this test")
@pytest.mark.usefixtures("bootstrap_test_motion_estimation2")
class TestMotionEstimationFromProjectedVolume:

    def test_180_with_return_projections(self, verbose=False):
        n_angles = 250
        cor = -10

        alpha_x = 4
        beta_x = 3
        alpha_y = -5
        beta_y = 10
        beta_z = 0
        orig_det_dist = 0

        angles0 = np.linspace(0, np.pi, n_angles, False)
        return_angles = np.deg2rad([180.0, 135.0, 90.0, 45.0, 0.0])
        angles = np.hstack([angles0, return_angles]).ravel()
        a = np.arange(angles0.size + return_angles.size) / angles0.size

        tx = _create_translations_vector(a, alpha_x, beta_x)
        ty = _create_translations_vector(a, alpha_y, beta_y)
        tz = _create_translations_vector(a, 0, beta_z)

        if not (hasattr(self, "volume")):
            self.volume = get_data("motion/mri_volume_subsampled.npy")
        sinos = project_volume(self.volume, angles, -tx, -ty, -tz, cor=-cor, orig_det_dist=orig_det_dist)
        radios = np.moveaxis(sinos, 1, 0)

        n_return_angles = return_angles.size
        projs_stack2 = radios[-n_return_angles:]
        outward_angles_indices = np.array([search_sorted(angles0, ra) for ra in return_angles])
        outward_angles = angles0[outward_angles_indices]
        projs_stack1 = radios[:-n_return_angles][outward_angles_indices]

        motion_estimator = MotionEstimation(
            projs_stack1,
            projs_stack2,
            outward_angles,
            return_angles,
            indices1=outward_angles_indices,
            indices2=np.arange(n_angles, n_angles + n_return_angles),
            shifts_estimator="DetectorTranslationAlongBeam",
        )
        motion_estimator.estimate_horizontal_motion(degree=2, cor=cor)
        motion_estimator.estimate_vertical_motion()

        gt_h = np.stack([-tx, ty], axis=1)
        gt_v = -tz
        if verbose:
            motion_estimator.plot_detector_shifts(cor=cor)
            motion_estimator.plot_movements(cor, angles_rad=angles0, gt_xy=gt_h[:n_angles], gt_z=gt_v[:n_angles])

        check_motion_estimation(
            motion_estimator,
            angles0,
            cor,
            gt_h,
            gt_v,
            fit_error_shifts_tol_vu=(0.01, 0.05),
            fit_error_det_tol_vu=(0.02, 0.5),
            fit_error_tol_xyz=(0.5, 1, 0.05),
            fit_error_det_all_angles_tol_vu=(0.02, 1),
        )

    def test_360(self, verbose=False):
        n_angles = 250
        cor = -5.5

        alpha_x = -2
        beta_x = 7.0
        alpha_y = -2
        beta_y = 3
        beta_z = 100
        orig_det_dist = 0

        angles = np.linspace(0, 2 * np.pi, n_angles, False)
        a = np.linspace(0, 1, angles.size, endpoint=False)  # theta/theta_max

        tx = _create_translations_vector(a, alpha_x, beta_x)
        ty = _create_translations_vector(a, alpha_y, beta_y)
        tz = _create_translations_vector(a, 0, beta_z)

        if not (hasattr(self, "volume")):
            self.volume = get_data("motion/mri_volume_subsampled.npy")

        sinos = project_volume(self.volume, angles, -tx, -ty, -tz, cor=-cor, orig_det_dist=orig_det_dist)
        radios = np.moveaxis(sinos, 1, 0)

        projs_stack1 = radios[: n_angles // 2]
        projs_stack2 = radios[n_angles // 2 :]
        angles1 = angles[: n_angles // 2]
        angles2 = angles[n_angles // 2 :]

        motion_estimator = MotionEstimation(
            projs_stack1, projs_stack2, angles1, angles2, shifts_estimator="phase_cross_correlation"
        )
        motion_estimator.estimate_horizontal_motion(degree=2, cor=cor)
        motion_estimator.estimate_vertical_motion()

        gt_xy = np.stack([-tx, ty], axis=1)
        gt_z = -tz

        if verbose:
            motion_estimator.plot_detector_shifts(cor=cor)
            motion_estimator.plot_movements(cor=cor, angles_rad=angles, gt_xy=gt_xy, gt_z=gt_z)

        check_motion_estimation(
            motion_estimator,
            angles,
            cor,
            gt_xy,
            gt_z,
            fit_error_shifts_tol_vu=(1e-5, 0.2),
            fit_error_det_tol_vu=(1e-5, 0.05),
            fit_error_tol_xyz=(0.1, 0.1, 1e-5),
            fit_error_det_all_angles_tol_vu=(1e-5, 0.1),
        )


if __name__ == "__main__":
    test1 = TestMotionEstimationFromProjectedVolume()
    test1.test_180_with_return_projections(verbose=True)
    test1.test_360(verbose=True)

    test2 = TestMotionEstimation()
    _bootstrap_test_motion_combined(test2)
    test2.test_estimate_motion_360(verbose=True)
    test2.test_estimate_motion_360_return(verbose=True)



from os import path
from multiprocessing.pool import ThreadPool

from ..utils import get_num_threads
from ..resources.nxflatfield import update_dataset_info_flats_darks
from ..resources.logger import LoggerOrPrint
from ..resources.dataset_analyzer import analyze_dataset
from ..preproc.ccd import CCDFilter
from ..pipeline.config_validators import convert_to_bool
from ..pipeline.estimators import TranslationsEstimator
from .utils import parse_params_values
from .cli_configs import EstimateMotionConfig


def estimate_motion():
    args = parse_params_values(
        EstimateMotionConfig,
        parser_description="Estimate sample motion and generate 'translation_movements_file' for nabu config file. ",
    )
    try:
        rot_center = float(args["rot_center"])
    except (ValueError, TypeError):
        rot_center = None

    logger = LoggerOrPrint(None)
    dataset_info = analyze_dataset(args["dataset"], logger=logger)
    do_ff = args["flatfield"]

    if do_ff:
        update_dataset_info_flats_darks(dataset_info, True, loading_mode="load_if_present")

    radios_filter = None
    do_ccd_filter = args["ccd_filter_size"] > 0
    if do_ccd_filter:
        radios_filter_obj = CCDFilter(
            dataset_info.radio_dims[::-1],
            kernel_size=args["ccd_filter_size"],
            median_clip_thresh=args["ccd_filter_threshold"],
        )

        def radios_filter(images):
            def _apply_median_clip(img):
                radios_filter_obj.median_clip_correction(img, output=img)

            with ThreadPool(get_num_threads()) as tp:
                tp.map(_apply_median_clip, images)

    est = TranslationsEstimator(
        dataset_info,
        do_flatfield=do_ff,
        rot_center=rot_center,
        angular_subsampling=args["subsampling"],
        deg_xy=args["deg_xy"],
        deg_z=args["deg_z"],
        shifts_estimator="phase_cross_correlation",
        radios_filter=radios_filter,
        extra_options={"window_size": args["win_size"]},
    )

    estimated_shifts_h, estimated_shifts_v, cor = est.estimate_motion()
    if convert_to_bool(args["verbose"]):
        err_vu = est.motion_estimator.get_max_fit_error(cor=rot_center)
        logger.info("Max fit error in 'u': %.2f pix \t\t Max fit error in 'v': %.2f pix" % (err_vu[1], err_vu[0]))
        est.motion_estimator.plot_detector_shifts(cor=rot_center)
        est.motion_estimator.plot_movements(cor=rot_center, angles_rad=dataset_info.rotation_angles)

    out_file = args["output_file"]
    est.generate_translations_movements_file(filename=out_file, only=args["only"] or None)
    logger.info(
        f"Wrote {out_file} - use 'translation_movements_file = {path.abspath(out_file)}' in nabu configuration file to correct for sample movements in the reconstruction'"
    )


if __name__ == "__main__":
    estimate_motion()



import os
from tempfile import TemporaryDirectory
import pytest
import numpy as np
from pint import get_application_registry
from nxtomo import NXtomo
from nabu.testutils import utilstest, __do_long_tests__, get_data
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer, analyze_dataset, ImageKey
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.utils import extract_parameters
from nabu.pipeline.estimators import CompositeCOREstimator, TranslationsEstimator
from nabu.pipeline.config import parse_nabu_config_file
from nabu.pipeline.estimators import SinoCORFinder, CORFinder

from nabu.estimation.tests.test_motion_estimation import (
    check_motion_estimation,
    project_volume,
    _create_translations_vector,
)


#
# Test CoR estimation with "composite-coarse-to-fine" (aka "near" in the legacy system vocable)
#


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    dataset_downloaded_path = utilstest.getfile("test_composite_cor_finder_data.h5")
    cls.theta_interval = 4.5 * 1  # this is given. Radios in the middle of steps 4.5 degree long
    # are set to zero for compression
    # You can still change it to a multiple of 4.5
    cls.cor_pix = 1321.625
    cls.abs_tol = 0.0001
    cls.dataset_info = HDF5DatasetAnalyzer(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.dataset_info, True)
    cls.cor_options = extract_parameters("side=300.0;  near_width = 20.0", sep=";")


@pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1 for this test")
@pytest.mark.usefixtures("bootstrap")
class TestCompositeCorFinder:
    def test(self):
        cor_finder = CompositeCOREstimator(
            self.dataset_info, theta_interval=self.theta_interval, cor_options=self.cor_options
        )

        cor_position = cor_finder.find_cor()
        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_pix
        assert np.isclose(self.cor_pix, cor_position, atol=self.abs_tol), message


@pytest.fixture(scope="class")
def bootstrap_bamboo_reduced(request):
    cls = request.cls
    cls.abs_tol = 0.2
    # Dataset without estimated_cor_frm_motor (non regression test)
    dataset_relpath = os.path.join("bamboo_reduced.nx")
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    conf_relpath = os.path.join("bamboo_reduced.conf")
    conf_downloaded_path = utilstest.getfile(conf_relpath)
    cls.ds_std = analyze_dataset(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.ds_std, True)
    cls.conf_std = parse_nabu_config_file(conf_downloaded_path)

    # Dataset with estimated_cor_frm_motor
    dataset_relpath = os.path.join("bamboo_reduced_bliss.nx")
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    conf_relpath = os.path.join("bamboo_reduced_bliss.conf")
    conf_downloaded_path = utilstest.getfile(conf_relpath)
    cls.ds_bliss = analyze_dataset(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.ds_bliss, True)
    cls.conf_bliss = parse_nabu_config_file(conf_downloaded_path)


@pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
@pytest.mark.usefixtures("bootstrap_bamboo_reduced")
class TestCorNearPos:
    # TODO adapt test file
    true_cor = 339.486 - 0.5

    def test_cor_sliding_standard(self):
        cor_options = extract_parameters(self.conf_std["reconstruction"].get("cor_options", None), sep=";")
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = CORFinder("sliding-window", self.ds_std, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor, cor, atol=self.abs_tol + 0.5), message  # FIXME

    def test_cor_fourier_angles_standard(self):
        cor_options = extract_parameters(self.conf_std["reconstruction"].get("cor_options", None), sep=";")
        # TODO modify test files
        if "near_pos" in cor_options and "near" in cor_options.get("side", "") == "near":
            cor_options["side"] = cor_options["near_pos"]
        #
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = SinoCORFinder("fourier-angles", self.ds_std, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor + 0.5, cor, atol=self.abs_tol), message

    def test_cor_sliding_bliss(self):
        cor_options = extract_parameters(self.conf_bliss["reconstruction"].get("cor_options", None), sep=";")
        # TODO modify test files
        if "near_pos" in cor_options and "near" in cor_options.get("side", "") == "near":
            cor_options["side"] = cor_options["near_pos"]
        #
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = CORFinder("sliding-window", self.ds_bliss, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor, cor, atol=self.abs_tol), message

    def test_cor_fourier_angles_bliss(self):
        cor_options = extract_parameters(self.conf_bliss["reconstruction"].get("cor_options", None), sep=";")
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = SinoCORFinder("fourier-angles", self.ds_bliss, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor + 0.5, cor, atol=self.abs_tol), message


def _add_fake_flats_and_dark_to_data(data, n_darks=10, n_flats=21, dark_val=1, flat_val=3):
    img_shape = data.shape[1:]
    # Use constant darks/flats, to avoid "reduction" (mean/median) issues
    fake_darks = np.ones((n_darks,) + img_shape, dtype=np.uint16) * dark_val
    fake_flats = np.ones((n_flats,) + img_shape, dtype=np.uint16) * flat_val
    return data * (fake_flats[0, 0, 0] - fake_darks[0, 0, 0]) + fake_darks[0, 0, 0], fake_darks, fake_flats


def _generate_nx_for_180_dataset(volume, output_file_path, n_darks=10, n_flats=21):

    n_angles = 250
    cor = -10

    alpha_x = 4
    beta_x = 3
    alpha_y = -5
    beta_y = 10
    beta_z = 0
    orig_det_dist = 0

    angles0 = np.linspace(0, np.pi, n_angles, False)
    return_angles = np.deg2rad([180.0, 135.0, 90.0, 45.0, 0.0])
    angles = np.hstack([angles0, return_angles]).ravel()
    a = np.arange(angles0.size + return_angles.size) / angles0.size

    tx = _create_translations_vector(a, alpha_x, beta_x)
    ty = _create_translations_vector(a, alpha_y, beta_y)
    tz = _create_translations_vector(a, 0, beta_z)

    sinos = project_volume(volume, angles, -tx, -ty, -tz, cor=-cor, orig_det_dist=orig_det_dist)
    data = np.moveaxis(sinos, 1, 0)

    sample_motion_xy = np.stack([-tx, ty], axis=1)
    sample_motion_z = -tz
    angles_deg = np.degrees(angles0)
    return_angles_deg = np.degrees(return_angles)
    n_return_radios = len(return_angles_deg)
    n_radios = data.shape[0] - n_return_radios

    ureg = get_application_registry()
    fake_raw_data, darks, flats = _add_fake_flats_and_dark_to_data(data, n_darks=n_darks, n_flats=n_flats)

    nxtomo = NXtomo()
    nxtomo.instrument.detector.data = np.concatenate(
        [
            darks,
            flats,
            fake_raw_data,  # radios + return radios (in float32 !)
        ]
    )
    image_key_control = np.concatenate(
        [
            [ImageKey.DARK_FIELD.value] * n_darks,
            [ImageKey.FLAT_FIELD.value] * n_flats,
            [ImageKey.PROJECTION.value] * n_radios,
            [ImageKey.ALIGNMENT.value] * n_return_radios,
        ]
    )
    nxtomo.instrument.detector.image_key_control = image_key_control

    rotation_angle = np.concatenate(
        [np.zeros(n_darks, dtype="f"), np.zeros(n_flats, dtype="f"), angles_deg, return_angles_deg]
    )
    nxtomo.sample.rotation_angle = rotation_angle * ureg.degree
    nxtomo.instrument.detector.field_of_view = "Full"
    nxtomo.instrument.detector.x_pixel_size = nxtomo.instrument.detector.y_pixel_size = 1 * ureg.micrometer
    nxtomo.save(file_path=output_file_path, data_path="entry", overwrite=True)

    return sample_motion_xy, sample_motion_z, cor


def _generate_nx_for_360_dataset(volume, output_file_path, n_darks=10, n_flats=21):

    n_angles = 250
    cor = -5.5

    alpha_x = -2
    beta_x = 7.0
    alpha_y = -2
    beta_y = 3
    beta_z = 100
    orig_det_dist = 0

    angles = np.linspace(0, 2 * np.pi, n_angles, False)
    a = np.linspace(0, 1, angles.size, endpoint=False)  # theta/theta_max

    tx = _create_translations_vector(a, alpha_x, beta_x)
    ty = _create_translations_vector(a, alpha_y, beta_y)
    tz = _create_translations_vector(a, 0, beta_z)

    sinos = project_volume(volume, angles, -tx, -ty, -tz, cor=-cor, orig_det_dist=orig_det_dist)
    data = np.moveaxis(sinos, 1, 0)

    sample_motion_xy = np.stack([-tx, ty], axis=1)
    sample_motion_z = -tz
    angles_deg = np.degrees(angles)

    ureg = get_application_registry()

    fake_raw_data, darks, flats = _add_fake_flats_and_dark_to_data(data, n_darks=n_darks, n_flats=n_flats)

    nxtomo = NXtomo()
    nxtomo.instrument.detector.data = np.concatenate([darks, flats, fake_raw_data])  # in float32 !

    image_key_control = np.concatenate(
        [
            [ImageKey.DARK_FIELD.value] * n_darks,
            [ImageKey.FLAT_FIELD.value] * n_flats,
            [ImageKey.PROJECTION.value] * data.shape[0],
        ]
    )
    nxtomo.instrument.detector.image_key_control = image_key_control

    rotation_angle = np.concatenate(
        [
            np.zeros(n_darks, dtype="f"),
            np.zeros(n_flats, dtype="f"),
            angles_deg,
        ]
    )
    nxtomo.sample.rotation_angle = rotation_angle * ureg.degree
    nxtomo.instrument.detector.field_of_view = "Full"
    nxtomo.instrument.detector.x_pixel_size = nxtomo.instrument.detector.y_pixel_size = 1 * ureg.micrometer
    nxtomo.save(file_path=output_file_path, data_path="entry", overwrite=True)

    return sample_motion_xy, sample_motion_z, cor


@pytest.fixture(scope="class")
def setup_test_motion_estimator(request):
    cls = request.cls
    cls.volume = get_data("motion/mri_volume_subsampled.npy")


@pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
@pytest.mark.usefixtures("setup_test_motion_estimator")
class TestMotionEstimator:

    def _setup(self, tmpdir):
        # pytest uses some weird data structure for "tmpdir"
        if not (isinstance(tmpdir, str)):
            tmpdir = str(tmpdir)
        #
        if getattr(self, "volume", None) is None:
            self.volume = get_data("motion/mri_volume_subsampled.npy")

    def test_estimate_motion_360_dataset(self, tmpdir, verbose=False):
        self._setup(tmpdir)
        nx_file_path = os.path.join(tmpdir, "mri_projected_360_motion.nx")
        sample_motion_xy, sample_motion_z, cor = _generate_nx_for_360_dataset(self.volume, nx_file_path)

        dataset_info = analyze_dataset(nx_file_path)

        translations_estimator = TranslationsEstimator(
            dataset_info, do_flatfield=True, rot_center=cor, angular_subsampling=5, deg_xy=2, deg_z=2
        )
        estimated_shifts_h, estimated_shifts_v, estimated_cor = translations_estimator.estimate_motion()

        s = translations_estimator.angular_subsampling
        if verbose:
            translations_estimator.motion_estimator.plot_detector_shifts(cor=cor)
            translations_estimator.motion_estimator.plot_movements(
                cor=cor,
                angles_rad=dataset_info.rotation_angles[::s],
                gt_xy=sample_motion_xy[::s, :],
                gt_z=sample_motion_z[::s],
            )
        check_motion_estimation(
            translations_estimator.motion_estimator,
            dataset_info.rotation_angles[::s],
            cor,
            sample_motion_xy[::s, :],
            sample_motion_z[::s],
            fit_error_shifts_tol_vu=(0.2, 0.2),
            fit_error_det_tol_vu=(1e-5, 5e-2),
            fit_error_tol_xyz=(0.05, 0.05, 0.05),
            fit_error_det_all_angles_tol_vu=(1e-5, 0.05),
        )

    def test_estimate_motion_180_dataset(self, tmpdir, verbose=False):
        self._setup(tmpdir)
        nx_file_path = os.path.join(tmpdir, "mri_projected_180_motion.nx")

        sample_motion_xy, sample_motion_z, cor = _generate_nx_for_180_dataset(self.volume, nx_file_path)

        dataset_info = analyze_dataset(nx_file_path)

        translations_estimator = TranslationsEstimator(
            dataset_info,
            do_flatfield=True,
            rot_center=cor,
            angular_subsampling=2,
            deg_xy=2,
            deg_z=2,
            shifts_estimator="DetectorTranslationAlongBeam",
        )
        estimated_shifts_h, estimated_shifts_v, estimated_cor = translations_estimator.estimate_motion()

        if verbose:
            translations_estimator.motion_estimator.plot_detector_shifts(cor=cor)
            translations_estimator.motion_estimator.plot_movements(
                cor=cor,
                angles_rad=dataset_info.rotation_angles,
                gt_xy=sample_motion_xy[: dataset_info.n_angles],
                gt_z=sample_motion_z[: dataset_info.n_angles],
            )

        check_motion_estimation(
            translations_estimator.motion_estimator,
            dataset_info.rotation_angles,
            cor,
            sample_motion_xy,
            sample_motion_z,
            fit_error_shifts_tol_vu=(0.02, 0.1),
            fit_error_det_tol_vu=(1e-2, 0.5),
            fit_error_tol_xyz=(0.5, 2, 1e-2),
            fit_error_det_all_angles_tol_vu=(1e-2, 2),
        )


if __name__ == "__main__":

    T = TestMotionEstimator()
    with TemporaryDirectory(suffix="_motion", prefix="nabu_testdata") as tmpdir:
        T.test_estimate_motion_360_dataset(tmpdir, verbose=True)
        T.test_estimate_motion_180_dataset(tmpdir, verbose=True)



from ..processing.unsharp import *  # noqa: F403
from ..utils import deprecation_warning

deprecation_warning("nabu.misc.unsharp has been moved to nabu.processing.unsharp", do_print=True, func_name="unsharp")



from nabu.utils import list_match_queries


def test_list_match_queries():

    # entry0000 .... entry0099
    avail = ["entry%04d" % i for i in range(100)]
    assert list_match_queries(avail, "entry0000") == ["entry0000"]
    assert list_match_queries(avail, ["entry0001"]) == ["entry0001"]
    assert list_match_queries(avail, ["entry000?"]) == ["entry%04d" % i for i in range(10)]
    assert list_match_queries(avail, ["entry*"]) == avail



import numpy as np


try:
    import corrct as cct

    __have_corrct__ = True
except ImportError:
    __have_corrct__ = False


class MLEMReconstructor:
    """
    A reconstructor for MLEM reconstruction using the CorrCT toolbox.

    Parameters
    ----------
    data_vwu_shape : tuple
        Shape of the input data, expected to be (n_slices, n_angles, n_dets). Raises an error if the shape is not 3D.
    angles_rad : numpy.ndarray
        Angles in radians for the projections. Must match the second dimension of `data_vwu_shape`.
    shifts_vu : numpy.ndarray, optional.
        Shifts in the v and u directions for each angle. If provided, must have the same number of cols as `angles_rad`. Each col is (tv,tu)
    cor : float, optional
        Center of rotation, which will be adjusted based on the sinogram width.
    n_iterations : int, optional
        Number of iterations for the MLEM algorithm. Default is 50.
    extra_options : dict, optional
        Additional options for the reconstruction process. Default options include:
        - scale_factor (float, default is 1.0): Scale factor for the reconstruction.
        - compute_shifts (boolean, default is False): Whether to compute shifts.
        - tomo_consistency (boolean, default is False): Whether to enforce tomographic consistency.
        - v_min_for_v_shifts (number, default is 0): Minimum value for vertical shifts.
        - v_max_for_v_shifts (number, default is None): Maximum value for vertical shifts.
        - v_min_for_u_shifts (number, default is 0): Minimum value for horizontal shifts.
        - v_max_for_u_shifts (number, default is None): Maximum value for horizontal shifts.
    """

    default_extra_options = {
        "compute_shifts": False,
        "tomo_consistency": False,
        "v_min_for_v_shifts": 0,
        "v_max_for_v_shifts": None,
        "v_min_for_u_shifts": 0,
        "v_max_for_u_shifts": None,
        "scale_factor": 1.0,
        "centered_axis": False,
        "clip_outer_circle": False,
        "outer_circle_value": 0.0,
        "filter_cutoff": 1.0,
        "padding_mode": None,
        "crop_filtered_data": True,
    }

    def __init__(
        self,
        data_vwu_shape,
        angles_rad,
        shifts_uv=None,
        cor=None,  # absolute
        n_iterations=50,
        extra_options=None,
    ):
        if not (__have_corrct__):
            raise ImportError("Need corrct package")
        self.angles_rad = angles_rad
        self.n_iterations = n_iterations
        self.scale_factor = extra_options.get("scale_factor", 1.0)

        self._configure_extra_options(extra_options)
        self._set_sino_shape(data_vwu_shape)
        self._set_shifts(shifts_uv, cor)

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _set_sino_shape(self, sinos_shape):
        if len(sinos_shape) != 3:
            raise ValueError("Expected a 3D shape")
        self.sinos_shape = sinos_shape
        self.n_sinos, self.n_angles, self.prj_width = sinos_shape
        if self.n_angles != len(self.angles_rad):
            raise ValueError(
                f"Number of angles ({len(self.angles_rad)}) does not match size of sinograms ({self.n_angles})."
            )

    def _set_shifts(self, shifts_uv, cor):
        if shifts_uv is None:
            self.shifts_vu = None
        else:
            if shifts_uv.shape[0] != self.n_angles:
                raise ValueError(
                    f"Number of shifts given ({shifts_uv.shape[0]}) does not mathc the number of projections ({self.n_angles})."
                )
            self.shifts_vu = -shifts_uv.copy().T[::-1]
        if cor is None:
            self.cor = 0.0
        else:
            self.cor = (
                -cor + (self.sinos_shape[-1] - 1) / 2.0
            )  # convert absolute to relative in the ASTRA convention, which is opposite to Nabu relative convention.

    def reset_rot_center(self, cor):
        self.cor = -cor + (self.sinos_shape[-1] - 1) / 2.0

    def reconstruct(self, data_vwu, x0=None):
        """
        data_align_vwu: numpy.ndarray or pycuda.gpuarray
            Raw data, with shape (n_sinograms, n_angles, width)
        output: optional
            Output array. If not provided, a new numpy array is returned
        """
        if not isinstance(data_vwu, np.ndarray):
            data_vwu = data_vwu.get()
        # data_vwu /= data_vwu.mean()

        # MLEM recons
        self.vol_geom_align = cct.models.VolumeGeometry.get_default_from_data(data_vwu)
        if self.shifts_vu is not None:
            self.prj_geom_align = cct.models.ProjectionGeometry.get_default_parallel()
            # Vertical shifts were handled in pipeline. Set them to ZERO
            self.shifts_vu[:, 0] = 0.0
            self.prj_geom_align.set_detector_shifts_vu(self.shifts_vu, self.cor)
        else:
            self.prj_geom_align = None

        variances_align = cct.processing.compute_variance_poisson(data_vwu)
        self.weights_align = cct.processing.compute_variance_weight(variances_align, normalized=True)  # , use_std=True
        self.data_term_align = cct.data_terms.DataFidelity_wl2(self.weights_align)
        solver = cct.solvers.MLEM(verbose=True, data_term=self.data_term_align)
        self.solver_opts = dict(lower_limit=0)  # , x_mask=cct.processing.circular_mask(vol_geom_align.shape_xyz[:-2])

        with cct.projectors.ProjectorUncorrected(
            self.vol_geom_align, self.angles_rad, rot_axis_shift_pix=self.cor, prj_geom=self.prj_geom_align
        ) as A:
            rec, _ = solver(A, data_vwu, iterations=self.n_iterations, x0=x0, **self.solver_opts)
        return rec * self.scale_factor



import os
import warnings
from functools import lru_cache
from multiprocessing import get_context
from multiprocessing.pool import Pool
from ..utils import BaseClassError, check_supported, no_decorator
from .fft_base import _BaseVKFFT

try:
    from pyvkfft.cuda import VkFFTApp as CudaVkFFTApp

    __has_vkfft__ = True
except (ImportError, OSError):
    __has_vkfft__ = False
    CudaVkFFTApp = BaseClassError
from ..cuda.processing import CudaProcessing

n_cached_ffts = int(os.getenv("NABU_FFT_CACHE", "0"))


maybe_cached = lru_cache(maxsize=n_cached_ffts) if n_cached_ffts > 0 else no_decorator


@maybe_cached
def _get_vkfft_cuda(*args, **kwargs):
    return CudaVkFFTApp(*args, **kwargs)


def get_vkfft_cuda(slf, *args, **kwargs):
    return _get_vkfft_cuda(*args, **kwargs)


class VKCUFFT(_BaseVKFFT):
    """
    Cuda FFT, using VKFFT backend
    """

    implem = "vkfft"
    backend = "cuda"
    ProcessingCls = CudaProcessing
    get_fft_obj = get_vkfft_cuda

    def _init_backend(self, backend_options):
        super()._init_backend(backend_options)
        self._vkfft_other_init_kwargs = {"stream": self.processing.stream}


def _has_vkfft(x):
    # should be run from within a Process
    try:
        from nabu.processing.fft_cuda import VKCUFFT, __has_vkfft__

        if not __has_vkfft__:
            return False
        _ = VKCUFFT((16,), "f")
        avail = True
    except (ImportError, RuntimeError, OSError, NameError):
        avail = False
    return avail


@lru_cache(maxsize=2)
def has_vkfft(safe=True):
    """
    Determine whether pyvkfft is available.
    For Cuda GPUs, vkfft relies on nvrtc which supports a narrow range of Cuda devices.
    Unfortunately, it's not possible to determine whether vkfft is available before creating a Cuda context.
    So we create a process (from scratch, i.e no fork), do the test within, and exit.
    This function cannot be tested from a notebook/console, a proper entry point has to be created (if __name__ == "__main__").
    """
    if not safe:
        return _has_vkfft(None)
    try:
        ctx = get_context("spawn")
        with Pool(1, context=ctx) as p:
            v = p.map(_has_vkfft, [1])[0]
    except AssertionError:
        # Can get AssertionError: daemonic processes are not allowed to have children
        # if the calling code is already a subprocess
        return _has_vkfft(None)
    return v


@lru_cache(maxsize=2)
def get_fft_class(backend="vkfft"):
    backends = {
        "vkfft": VKCUFFT,
        "pyvkfft": VKCUFFT,
    }

    def get_fft_cls(asked_fft_backend):
        asked_fft_backend = asked_fft_backend.lower()
        check_supported(asked_fft_backend, list(backends.keys()), "Cuda FFT backend name")
        return backends[asked_fft_backend]

    asked_fft_backend_env = os.environ.get("NABU_FFT_BACKEND", "")
    if asked_fft_backend_env != "":
        return get_fft_cls(asked_fft_backend_env)

    avail_fft_implems = get_available_fft_implems()
    if len(avail_fft_implems) == 0:
        raise RuntimeError("Could not any Cuda FFT implementation. Please install pyvkfft")
    if backend not in avail_fft_implems:
        warnings.warn("Could not get FFT backend '%s'" % backend, RuntimeWarning)
        backend = avail_fft_implems[0]

    return get_fft_cls(backend)


@lru_cache(maxsize=1)
def get_available_fft_implems():
    avail_implems = []
    if has_vkfft(safe=True):
        avail_implems.append("vkfft")
    return avail_implems



import pytest
import numpy as np
from nabu.testutils import get_data

from nabu.cuda.utils import __has_pycuda__
from nabu.reconstruction.mlem import MLEMReconstructor, __have_corrct__


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    datafile = get_data("test_mlem.npz")
    cls.data_wvu = datafile["data_wvu"]
    cls.angles_rad = datafile["angles_rad"]
    cls.pixel_size_cm = datafile["pixel_size"] * 1e4  # pixel_size originally in um
    cls.true_cor = datafile["true_cor"]
    cls.mlem_cor_None_nosh = datafile["mlem_cor_None_nosh"]
    cls.mlem_cor_truecor_nosh = datafile["mlem_cor_truecor_nosh"]
    cls.mlem_cor_truecor_shifts_v0 = datafile["mlem_cor_truecor_shifts_v0"]
    cls.shifts_uv_v0 = datafile["shifts_uv_v0"]
    cls.shifts_uv = datafile["shifts_uv"]

    cls.tol = 1.3e-4


@pytest.mark.skipif(not (__has_pycuda__ and __have_corrct__), reason="Need pycuda and corrct for this test")
@pytest.mark.usefixtures("bootstrap")
class TestMLEMReconstructor:
    """These tests test the general MLEM reconstruction algorithm
    and the behavior of the reconstruction with respect to horizontal shifts.
    Only horizontal shifts are tested here because vertical shifts are handled outside
    the reconstruction object, but in the embedding reconstruction pipeline. See FullFieldReconstructor
    It is compared against a reference reconstruction generated with the `rec_mlem` function
    defined in the `generate_test_data.py` script.
    """

    def _rec_mlem(self, cor, shifts_uv, data_wvu, angles_rad):
        n_angles, n_z, n_x = data_wvu.shape

        mlem = MLEMReconstructor(
            (n_z, n_angles, n_x),
            angles_rad,
            shifts_uv=shifts_uv,
            cor=cor,
            n_iterations=50,
            extra_options={"centered_axis": True, "clip_outer_circle": True, "scale_factor": 1 / self.pixel_size_cm},
        )
        rec_mlem = mlem.reconstruct(data_wvu.swapaxes(0, 1))
        return rec_mlem

    def test_simple_mlem_recons_cor_None_nosh(self):
        slice_index = 25
        rec = self._rec_mlem(None, None, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_None_nosh)
        assert np.max(delta) < self.tol

    def test_simple_mlem_recons_cor_truecor_nosh(self):
        slice_index = 25
        rec = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_truecor_nosh)
        assert np.max(delta) < 2.6e-4

    def test_compare_with_fbp(self):
        from nabu.reconstruction.fbp import Backprojector

        def _rec_fbp(cor, shifts_uv, data_wvu, angles_rad):
            n_angles, n_z, n_x = data_wvu.shape

            if shifts_uv is None:
                fbp = Backprojector(
                    (n_angles, n_x),
                    angles=angles_rad,
                    rot_center=cor,
                    halftomo=False,
                    padding_mode="edges",
                    extra_options={
                        "centered_axis": True,
                        "clip_outer_circle": True,
                        "scale_factor": 1 / self.pixel_size_cm,
                    },
                )
            else:
                fbp = Backprojector(
                    (n_angles, n_x),
                    angles=angles_rad,
                    rot_center=cor,
                    halftomo=False,
                    padding_mode="edges",
                    extra_options={
                        "centered_axis": True,
                        "clip_outer_circle": True,
                        "scale_factor": 1 / self.pixel_size_cm,  # convert um to cm
                        "axis_correction": shifts_uv[:, 0],
                    },
                )

            rec_fbp = np.zeros((n_z, n_x, n_x), "f")
            for i in range(n_z):
                rec_fbp[i] = fbp.fbp(data_wvu[:, i])

            return rec_fbp

        fbp = _rec_fbp(self.true_cor, None, self.data_wvu, self.angles_rad)[25]
        mlem = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)[25]
        delta = np.abs(fbp - mlem)
        assert (
            np.max(delta) < 400
        )  # These two should not be really equal. But the test should test that both algo FBP and MLEM behave similarly.

    def test_mlem_zeroshifts_equal_noshifts(self):
        shifts = np.zeros((len(self.angles_rad), 2))
        rec_nosh = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)
        rec_zerosh = self._rec_mlem(self.true_cor, shifts, self.data_wvu, self.angles_rad)
        delta = np.abs(rec_nosh - rec_zerosh)
        assert np.max(delta) < self.tol

    def test_mlem_recons_with_u_shifts(self):
        slice_index = 25
        rec = self._rec_mlem(self.true_cor, self.shifts_uv_v0, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_truecor_shifts_v0)
        assert np.max(delta) < self.tol



from ..processing.fft_cuda import *  # noqa: F403
from ..utils import deprecation_warning

deprecation_warning("nabu.cuda.fft has been moved to nabu.processing.fft_cuda", do_print=True, func_name="fft_cuda")



from functools import lru_cache
import os
from multiprocessing import get_context
from multiprocessing.pool import Pool

from ..utils import BaseClassError, no_decorator
from .fft_base import _BaseVKFFT
from ..opencl.processing import OpenCLProcessing

try:
    from pyvkfft.opencl import VkFFTApp as OpenCLVkFFTApp

    __has_vkfft__ = True
except (ImportError, OSError):
    __has_vkfft__ = False
    vk_clfft = None
    OpenCLVkFFTApp = BaseClassError

n_cached_ffts = int(os.getenv("NABU_FFT_CACHE", "0"))
maybe_cached = lru_cache(maxsize=n_cached_ffts) if n_cached_ffts > 0 else no_decorator


@maybe_cached
def _get_vkfft_opencl(*args, **kwargs):
    return OpenCLVkFFTApp(*args, **kwargs)


def get_vkfft_opencl(slf, *args, **kwargs):
    return _get_vkfft_opencl(*args, **kwargs)


class VKCLFFT(_BaseVKFFT):
    """
    OpenCL FFT, using VKFFT backend
    """

    implem = "vkfft"
    backend = "opencl"
    ProcessingCls = OpenCLProcessing
    get_fft_obj = get_vkfft_opencl

    def _init_backend(self, backend_options):
        super()._init_backend(backend_options)
        self._vkfft_other_init_kwargs = {"queue": self.processing.queue}


def _has_vkfft(x):
    # should be run from within a Process
    try:
        from nabu.processing.fft_opencl import VKCLFFT, __has_vkfft__

        if not __has_vkfft__:
            return False
        _ = VKCLFFT((16,), "f")
        avail = True
    except (RuntimeError, OSError):
        avail = False
    return avail


def has_vkfft(safe=True):
    """
    Determine whether pyvkfft is available.
    This function cannot be tested from a notebook/console, a proper entry point has to be created (if __name__ == "__main__").
    """
    if not safe:
        return _has_vkfft(None)
    try:
        ctx = get_context("spawn")
        with Pool(1, context=ctx) as p:
            v = p.map(_has_vkfft, [1])[0]
    except AssertionError:
        # Can get AssertionError: daemonic processes are not allowed to have children
        # if the calling code is already a subprocess
        return _has_vkfft(None)
    return v



import numpy as np


class IterativeBase:

    backend = None  # placeholder
    implementation = None  # placeholder

    default_extra_options = {
        "axis_correction": None,
        "centered_axis": False,
        "clip_outer_circle": False,
        "scale_factor": None,
        "outer_circle_value": 0.0,
    }

    backend_processing_class = ProcessingBase

    def __init__(
        self,
        sino_shape,
        angles=None,
        rot_center=None,
        halftomo=False,
        filter_name=None,
        slice_roi=None,
        scale_factor=None,
        extra_options=None,
        backend_options=None,
    ): ...



import numpy as np
from ..cuda.processing import CudaProcessing
from ..utils import docstring, get_cuda_srcfile
from ..processing.padding_cuda import CudaPadding
from ..processing.fft_cuda import get_fft_class
from .filtering import SinoFilter


class CudaSinoFilter(SinoFilter):
    default_extra_options = {**SinoFilter.default_extra_options, "fft_backend": "vkfft"}

    @docstring(SinoFilter)
    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        crop_filtered_data=True,
        extra_options=None,
        cuda_options=None,
    ):
        self._cuda_options = cuda_options or {}
        self.cuda = CudaProcessing(**self._cuda_options)
        super().__init__(
            sino_shape,
            filter_name=filter_name,
            padding_mode=padding_mode,
            crop_filtered_data=crop_filtered_data,
            extra_options=extra_options,
        )
        self._init_kernels()

    def _init_fft(self):
        fft_cls = get_fft_class(self.extra_options["fft_backend"])
        self.fft = fft_cls(
            self.sino_padded_shape,
            dtype=np.float32,
            axes=(-1,),
        )

    def _allocate_memory(self):
        self.d_filter_f = self.cuda.allocate_array("d_filter_f", (self.sino_f_shape[-1],), dtype=np.complex64)
        self.d_sino_padded = self.cuda.allocate_array("d_sino_padded", self.fft.shape)
        self.d_sino_f = self.cuda.allocate_array("d_sino_f", self.fft.shape_out, dtype=np.complex64)

    def set_filter(self, h_filt, normalize=True):
        super().set_filter(h_filt, normalize=normalize)
        self.d_filter_f[:] = self.filter_f[:]

    def _init_kernels(self):
        # pointwise complex multiplication
        fname = get_cuda_srcfile("ElementOp.cu")
        if self.ndim == 2:
            kernel_name = "inplace_complex_mul_2Dby1D"
            kernel_sig = "PPii"
        else:
            kernel_name = "inplace_complex_mul_3Dby1D"
            kernel_sig = "PPiii"
        self.mult_kernel = self.cuda.kernel(kernel_name, filename=fname, signature=kernel_sig)
        self.kern_args = (self.d_sino_f, self.d_filter_f)
        self.kern_args += self.d_sino_f.shape[::-1]
        # padding
        self.padding_kernel = CudaPadding(
            self.sino_shape,
            ((0, 0), (self.pad_left, self.pad_right)),
            mode=self.padding_mode,
            cuda_options=self._cuda_options,
        )

    def filter_sino(self, sino, output=None):
        self._check_array(sino)
        if not (isinstance(sino, self.cuda.array_class)):
            sino = self.cuda.set_array("sino", sino)
        elif not (sino.flags.c_contiguous):
            # Transfer the device array into another, c-contiguous, device array
            # We can throw an error as well in this case, but often we do something like fbp(radios[:, i, :])
            sino_tmp = self.cuda.allocate_array("sino_contig", sino.shape)
            sino_tmp.set(sino)
            sino = sino_tmp

        # Padding
        self.padding_kernel(sino, output=self.d_sino_padded)

        # FFT
        self.fft.fft(self.d_sino_padded, output=self.d_sino_f)

        # multiply padded sinogram with filter in the Fourier domain
        self.mult_kernel(*self.kern_args)  # TODO tune block size ?

        # iFFT
        self.fft.ifft(self.d_sino_f, output=self.d_sino_padded)

        # return
        if output is None:
            if not (self.crop_filtered_data):
                # No need to allocate extra memory here
                return self.d_sino_padded
            res = self.cuda.allocate_array("output", self.output_shape)
        else:
            res = output

        if self.crop_filtered_data:
            # res[:] = sino_filtered[..., : self.dwidth]  # pylint: disable=E1126 # ?!
            res[:] = self.d_sino_padded[..., self.pad_left : -self.pad_right]  # pylint: disable=E1126 # ?!
        else:
            res[:] = self.d_sino_padded[:]

        return res

    __call__ = filter_sino
